# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for kfp.dsl.types.type_annotations."""

import sys
from typing import Any, Dict, List, Optional, Union
import unittest

from absl.testing import parameterized
from kfp import dsl
from kfp.dsl import Input
from kfp.dsl import Output
from kfp.dsl.types import artifact_types
from kfp.dsl.types import type_annotations
from kfp.dsl.types.artifact_types import Model
from kfp.dsl.types.type_annotations import InputAnnotation
from kfp.dsl.types.type_annotations import InputPath
from kfp.dsl.types.type_annotations import OutputAnnotation
from kfp.dsl.types.type_annotations import OutputPath


class AnnotationsTest(parameterized.TestCase):

    @parameterized.parameters([
        Input[Model],
        Output[Model],
        Output[List[Model]],
        Output['MyArtifact'],
    ])
    def test_is_artifact_annotation(self, annotation):
        self.assertTrue(
            type_annotations.is_Input_Output_artifact_annotation(annotation))

    @parameterized.parameters([
        Model,
        int,
        'Dataset',
        List[str],
        Optional[str],
    ])
    def test_is_not_artifact_annotation(self, annotation):
        self.assertFalse(
            type_annotations.is_Input_Output_artifact_annotation(annotation))

    @parameterized.parameters([
        Input[Model],
        Input,
    ])
    def test_is_input_artifact(self, annotation):
        self.assertTrue(
            type_annotations.is_artifact_wrapped_in_Input(annotation))

    @parameterized.parameters([
        Output[Model],
        Output,
    ])
    def test_is_not_input_artifact(self, annotation):
        self.assertFalse(
            type_annotations.is_artifact_wrapped_in_Input(annotation))

    @parameterized.parameters([
        Output[Model],
        Output[List[Model]],
    ])
    def test_is_output_artifact(self, annotation):
        self.assertTrue(
            type_annotations.is_artifact_wrapped_in_Output(annotation))

    @parameterized.parameters([
        Input[Model],
        Input[List[Model]],
        Input,
    ])
    def test_is_not_output_artifact(self, annotation):
        self.assertFalse(
            type_annotations.is_artifact_wrapped_in_Output(annotation))

    def test_get_io_artifact_class(self):
        self.assertEqual(
            type_annotations.get_io_artifact_class(Output[Model]), Model)
        self.assertEqual(
            type_annotations.get_io_artifact_class(Output[List[Model]]), Model)
        self.assertEqual(
            type_annotations.get_io_artifact_class(Input[List[Model]]), Model)

        self.assertEqual(type_annotations.get_io_artifact_class(Input), None)
        self.assertEqual(type_annotations.get_io_artifact_class(Output), None)
        self.assertEqual(type_annotations.get_io_artifact_class(Model), None)
        self.assertEqual(type_annotations.get_io_artifact_class(str), None)

    def test_get_io_artifact_annotation(self):
        self.assertEqual(
            type_annotations.get_input_or_output_marker(Output[Model]),
            OutputAnnotation)
        self.assertEqual(
            type_annotations.get_input_or_output_marker(Output[List[Model]]),
            OutputAnnotation)
        self.assertEqual(
            type_annotations.get_input_or_output_marker(Input[Model]),
            InputAnnotation)
        self.assertEqual(
            type_annotations.get_input_or_output_marker(Input[List[Model]]),
            InputAnnotation)
        self.assertEqual(
            type_annotations.get_input_or_output_marker(Input), InputAnnotation)
        self.assertEqual(
            type_annotations.get_input_or_output_marker(Output),
            OutputAnnotation)

        self.assertEqual(
            type_annotations.get_input_or_output_marker(Model), None)
        self.assertEqual(type_annotations.get_input_or_output_marker(str), None)

    @parameterized.parameters(
        {
            'original_annotation': str,
            'expected_annotation': str,
        },
        {
            'original_annotation': 'MyCustomType',
            'expected_annotation': 'MyCustomType',
        },
        {
            'original_annotation': List[int],
            'expected_annotation': List[int],
        },
        {
            'original_annotation': Optional[str],
            'expected_annotation': str,
        },
        {
            'original_annotation': Optional[Dict[str, float]],
            'expected_annotation': Dict[str, float],
        },
        {
            'original_annotation': Optional[List[Dict[str, Any]]],
            'expected_annotation': List[Dict[str, Any]],
        },
        {
            'original_annotation': Input[Model],
            'expected_annotation': Input[Model],
        },
        {
            'original_annotation': InputPath('Model'),
            'expected_annotation': InputPath('Model'),
        },
        {
            'original_annotation': OutputPath(Model),
            'expected_annotation': OutputPath(Model),
        },
    )
    def test_maybe_strip_optional_from_annotation(self, original_annotation,
                                                  expected_annotation):
        self.assertEqual(
            expected_annotation,
            type_annotations.maybe_strip_optional_from_annotation(
                original_annotation))

    @parameterized.parameters(
        {
            'original_type_name': 'str',
            'expected_type_name': 'str',
        },
        {
            'original_type_name': 'typing.List[int]',
            'expected_type_name': 'List',
        },
        {
            'original_type_name': 'List[int]',
            'expected_type_name': 'List',
        },
        {
            'original_type_name': 'Dict[str, str]',
            'expected_type_name': 'Dict',
        },
        {
            'original_type_name': 'List[Dict[str, str]]',
            'expected_type_name': 'List',
        },
    )
    def test_get_short_type_name(self, original_type_name, expected_type_name):
        self.assertEqual(
            expected_type_name,
            type_annotations.get_short_type_name(original_type_name))


class TestIsArtifact(parameterized.TestCase):

    @parameterized.parameters([{
        'obj': obj
    } for obj in artifact_types._SCHEMA_TITLE_TO_TYPE.values()])
    def test_true_class(self, obj):
        self.assertTrue(type_annotations.is_artifact_class(obj))

    @parameterized.parameters([{
        'obj': obj(name='name', uri='uri', metadata={})
    } for obj in artifact_types._SCHEMA_TITLE_TO_TYPE.values()])
    def test_true_instance(self, obj):
        self.assertTrue(type_annotations.is_artifact_class(obj))

    @parameterized.parameters([{'obj': 'string'}, {'obj': 1}, {'obj': int}])
    def test_false(self, obj):
        self.assertFalse(type_annotations.is_artifact_class(obj))

    def test_false_no_schema_title(self):

        class MissingSchemaTitle:
            schema_version = ''

        self.assertFalse(type_annotations.is_artifact_class(MissingSchemaTitle))

    def test_false_no_schema_version(self):

        class MissingSchemaVersion:
            schema_title = ''

        self.assertFalse(
            type_annotations.is_artifact_class(MissingSchemaVersion))


class ArtifactSubclass(dsl.Artifact):
    pass


class NotArtifactSubclass:
    pass


class TestIsSubclassOfArtifact(parameterized.TestCase):

    @parameterized.parameters([{
        'obj': obj
    } for obj in [
        dsl.Artifact,
        dsl.Dataset,
        dsl.Metrics,
        ArtifactSubclass,
    ]])
    def test_true(self, obj):
        self.assertTrue(type_annotations.issubclass_of_artifact(obj))

    @parameterized.parameters([{
        'obj': obj
    } for obj in [
        dsl.Artifact(),
        dsl.Dataset(),
        1,
        NotArtifactSubclass,
    ]])
    def test_false(self, obj):
        self.assertFalse(type_annotations.issubclass_of_artifact(obj))


class TestIsGenericList(parameterized.TestCase):

    @parameterized.parameters([{
        'obj': obj
    } for obj in [
        List,
        List[str],
        List[dsl.Artifact],
        List[Dict[str, str]],
    ] + ([
        list,
        list[str],
    ] if sys.version_info >= (3, 9, 0) else [])])
    def test_true(self, obj):
        self.assertTrue(type_annotations.is_generic_list(obj))

    @parameterized.parameters([{
        'obj': obj
    } for obj in [
        Optional[List[str]],
        Dict[str, str],
        str,
        int,
        dsl.Artifact,
    ]])
    def test_false(self, obj):
        self.assertFalse(type_annotations.is_generic_list(obj))


class TestGetInnerType(parameterized.TestCase):

    @parameterized.parameters([{
        'annotation': annotation,
        'expected': expected
    } for annotation, expected in [
        (int, None),
        (Optional[int], (int, type(None))),
        (Union[int, None], (int, type(None))),
        (List[str], str),
        (Dict[str, str], (str, str)),
        (List[dsl.Artifact], dsl.Artifact),
    ] + ([
        (list[str], str),
        (dict[str, str], (str, str)),
        (list[dsl.Artifact], dsl.Artifact),
    ] if sys.version_info >= (3, 9, 0) else [])])
    def test(self, annotation, expected):
        actual = type_annotations.get_inner_type(annotation)
        self.assertEqual(actual, expected)


if __name__ == '__main__':
    unittest.main()



# Copyright 2023 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilitites for formatting, coloring, and controlling the output of logs."""
import builtins
import contextlib
import datetime
import logging
import shutil
import sys
from typing import Any, Dict, Generator, List

from kfp import dsl
from kfp.local import status


class Color:
    # color for task name
    CYAN = '\033[96m'
    # color for status success
    GREEN = '\033[92m'
    # color for status failure
    RED = '\033[91m'
    # color for pipeline name
    MAGENTA = '\033[95m'
    RESET = '\033[0m'


class MillisecondFormatter(logging.Formatter):

    def formatTime(
        self,
        record: logging.LogRecord,
        datefmt: str = None,
    ) -> str:
        created = datetime.datetime.fromtimestamp(record.created)
        s = created.strftime(datefmt)
        # truncate microseconds to milliseconds
        return s[:-3]


@contextlib.contextmanager
def local_logger_context() -> Generator[None, None, None]:
    """Context manager for creating and reseting the local execution logger."""

    logger = logging.getLogger()
    original_level = logger.level
    original_handlers = logger.handlers[:]
    formatter = MillisecondFormatter(
        fmt='%(asctime)s - %(levelname)s - %(message)s',
        datefmt='%H:%M:%S.%f',
    )
    # use sys.stdout so that both inner process and outer process logs
    # go to stdout
    # this is needed for logs to present sequentially in a colab notebook,
    # since stderr will print above stdout
    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(formatter)
    logger.handlers.clear()
    logger.addHandler(handler)
    logger.setLevel(logging.INFO)

    try:
        yield
    finally:
        logger.setLevel(original_level)
        logger.handlers.clear()
        for handler in original_handlers:
            logger.addHandler(handler)


@contextlib.contextmanager
def indented_print(num_spaces: int = 4) -> Generator[None, None, None]:
    """Context manager to indent all print statements in its scope by
    num_prints.

    Useful for visually separating a subprocess logs from the outer
    process logs.
    """
    original_print = builtins.print

    def indented_print_function(*args, **kwargs):
        original_print(' ' * num_spaces, end='')
        return original_print(*args, **kwargs)

    builtins.print = indented_print_function
    try:
        yield
    finally:
        builtins.print = original_print


def color_text(text: str, color: Color) -> str:
    return f'{color}{text}{Color.RESET}'


def make_log_lines_for_artifact(artifact: dsl.Artifact,) -> List[str]:
    """Returns a list of log lines that represent a single artifact output."""
    artifact_class_name_and_paren = f'{artifact.__class__.__name__}( '
    # name
    artifact_lines = [f"{artifact_class_name_and_paren}name='{artifact.name}',"]
    newline_spaces = len(artifact_class_name_and_paren) * ' '
    # uri
    artifact_lines.append(f"{newline_spaces}uri='{artifact.uri}',")
    # metadata
    artifact_lines.append(f'{newline_spaces}metadata={artifact.metadata} )')
    return artifact_lines


def make_log_lines_for_outputs(outputs: Dict[str, Any]) -> List[str]:
    """Returns a list of log lines to repesent the outputs of a task."""
    INDENT = ' ' * 4
    SEPARATOR = ': '
    output_lines = []
    for key, value in outputs.items():
        key_chars = INDENT + key + SEPARATOR

        # present artifacts
        if isinstance(value, dsl.Artifact):
            artifact_lines = make_log_lines_for_artifact(value)

            first_artifact_line = artifact_lines[0]
            output_lines.append(f'{key_chars}{first_artifact_line}')

            remaining_artifact_lines = artifact_lines[1:]
            # indent to align with first char in artifact
            # to visually separate output keys
            remaining_artifact_lines = [
                len(key_chars) * ' ' + l for l in remaining_artifact_lines
            ]
            output_lines.extend(remaining_artifact_lines)

        # present params
        else:
            value = f"'{value}'" if isinstance(value, str) else value
            output_lines.append(f'{key_chars}{value}')

    return output_lines


def print_horizontal_line() -> None:
    columns, _ = shutil.get_terminal_size(fallback=(80, 24))
    print('-' * columns)


def format_task_name(task_name: str) -> str:
    return color_text(f'{task_name!r}', Color.CYAN)


def format_status(task_status: status.Status) -> str:
    if task_status == status.Status.SUCCESS:
        return color_text(task_status.name, Color.GREEN)
    elif task_status == status.Status.FAILURE:
        return color_text(task_status.name, Color.RED)
    else:
        raise ValueError(f'Got unknown status: {task_status}')


def format_pipeline_name(pipeline_name: str) -> str:
    return color_text(f'{pipeline_name!r}', Color.MAGENTA)



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import inspect
import json
import os
import re
from typing import Any, Callable, Dict, List, Optional, Union
import warnings

from kfp import dsl
from kfp.dsl import task_final_status
from kfp.dsl.types import artifact_types
from kfp.dsl.types import type_annotations


class Executor:
    """Executor executes Python function components."""

    def __init__(
        self,
        executor_input: Dict,
        function_to_execute: Union[Callable,
                                   'python_component.PythonComponent'],
    ):

        if hasattr(function_to_execute, 'python_func'):
            self.func = function_to_execute.python_func
        else:
            self.func = function_to_execute

        self.executor_input = executor_input
        self.executor_output_path = self.executor_input['outputs']['outputFile']

        # drop executor_output.json part from the outputFile path
        artifact_types.CONTAINER_TASK_ROOT = os.path.split(
            self.executor_output_path)[0]

        self.input_artifacts: Dict[str, Union[dsl.Artifact,
                                              List[dsl.Artifact]]] = {}
        self.output_artifacts: Dict[str, dsl.Artifact] = {}
        self.assign_input_and_output_artifacts()

        self.return_annotation = inspect.signature(self.func).return_annotation
        self.excutor_output = {}

    def assign_input_and_output_artifacts(self) -> None:
        for name, artifacts in self.executor_input.get('inputs',
                                                       {}).get('artifacts',
                                                               {}).items():
            list_of_artifact_proto_structs = artifacts.get('artifacts')
            if list_of_artifact_proto_structs:
                annotation = self.func.__annotations__[name]
                # InputPath has no attribute __origin__ and also should be handled as a single artifact
                annotation = type_annotations.maybe_strip_optional_from_annotation(
                    annotation)
                is_list_of_artifacts = (
                    type_annotations.is_Input_Output_artifact_annotation(
                        annotation) and
                    type_annotations.is_list_of_artifacts(annotation.__origin__)
                ) or type_annotations.is_list_of_artifacts(annotation)
                if is_list_of_artifacts:
                    # Get the annotation of the inner type of the list
                    # to use when creating the artifacts
                    inner_annotation = type_annotations.get_inner_type(
                        annotation)

                    self.input_artifacts[name] = [
                        self.make_artifact(
                            msg,
                            name,
                            self.func,
                            annotation=inner_annotation,
                        ) for msg in list_of_artifact_proto_structs
                    ]
                else:
                    self.input_artifacts[name] = self.make_artifact(
                        list_of_artifact_proto_structs[0],
                        name,
                        self.func,
                    )

        for name, artifacts in self.executor_input.get('outputs',
                                                       {}).get('artifacts',
                                                               {}).items():
            list_of_artifact_proto_structs = artifacts.get('artifacts')
            if list_of_artifact_proto_structs:
                output_artifact = self.make_artifact(
                    list_of_artifact_proto_structs[0],
                    name,
                    self.func,
                )
                self.output_artifacts[name] = output_artifact
                makedirs_recursively(output_artifact.path)

    def make_artifact(
        self,
        runtime_artifact: Dict,
        name: str,
        func: Callable,
        annotation: Optional[Any] = None,
    ) -> Any:
        annotation = func.__annotations__.get(
            name) if annotation is None else annotation
        if isinstance(annotation, type_annotations.InputPath):
            schema_title, _ = annotation.type.split('@')
            if schema_title in artifact_types._SCHEMA_TITLE_TO_TYPE:
                artifact_cls = artifact_types._SCHEMA_TITLE_TO_TYPE[
                    schema_title]
            else:
                raise TypeError(
                    f'Invalid type argument to {type_annotations.InputPath.__name__}: {annotation.type}'
                )
        else:
            artifact_cls = annotation
        return create_artifact_instance(
            runtime_artifact, fallback_artifact_cls=artifact_cls)

    def get_input_artifact(self, name: str) -> Optional[dsl.Artifact]:
        return self.input_artifacts.get(name)

    def get_output_artifact(self, name: str) -> Optional[dsl.Artifact]:
        return self.output_artifacts.get(name)

    def get_input_parameter_value(self, parameter_name: str) -> Optional[str]:
        parameter_values = self.executor_input.get('inputs', {}).get(
            'parameterValues', None)

        if parameter_values is not None:
            return parameter_values.get(parameter_name, None)

        return None

    def get_output_parameter_path(self, parameter_name: str) -> Optional[str]:
        parameter = self.executor_input.get('outputs', {}).get(
            'parameters', {}).get(parameter_name, None)
        if parameter is None:
            return None

        path = parameter.get('outputFile', None)
        if path:
            makedirs_recursively(path)
        return path

    def get_output_artifact_path(self, artifact_name: str) -> str:
        output_artifact = self.output_artifacts.get(artifact_name)
        if not output_artifact:
            raise ValueError(
                f'Failed to get output artifact path for artifact name {artifact_name}'
            )
        return output_artifact.path

    def get_input_artifact_path(self, artifact_name: str) -> str:
        input_artifact = self.input_artifacts.get(artifact_name)
        if not input_artifact:
            raise ValueError(
                f'Failed to get input artifact path for artifact name {artifact_name}'
            )
        return input_artifact.path

    def write_output_parameter_value(
            self, name: str, value: Union[str, int, float, bool, dict, list,
                                          Dict, List]) -> None:
        if isinstance(value, (float, int)):
            output = str(value)
        elif isinstance(value, str):
            # value is already a string.
            output = value
        elif isinstance(value, (bool, list, dict)):
            output = json.dumps(value)
        else:
            raise ValueError(
                f'Unable to serialize unknown type `{value}` for parameter input with value `{type(value)}`'
            )

        if not self.excutor_output.get('parameterValues'):
            self.excutor_output['parameterValues'] = {}

        self.excutor_output['parameterValues'][name] = value

    def write_output_artifact_payload(self, name: str, value: Any) -> None:
        path = self.get_output_artifact_path(name)
        with open(path, 'w') as f:
            f.write(str(value))

    def handle_single_return_value(self, output_name: str, annotation_type: Any,
                                   return_value: Any) -> None:
        if is_parameter(annotation_type):
            origin_type = getattr(annotation_type, '__origin__',
                                  None) or annotation_type
            # relax float-typed return to allow both int and float.
            if origin_type == float:
                accepted_types = (int, float)
            # TODO: relax str-typed return to allow all primitive types?
            else:
                accepted_types = origin_type
            if not isinstance(return_value, accepted_types):
                raise ValueError(
                    f'Function `{self.func.__name__}` returned value of type {type(return_value)}; want type {origin_type}'
                )
            self.write_output_parameter_value(output_name, return_value)

        elif is_artifact(annotation_type):
            if isinstance(return_value, artifact_types.Artifact):
                # for -> Artifact annotations, where the user returns an artifact
                artifact_name = self.executor_input['outputs']['artifacts'][
                    output_name]['artifacts'][0]['name']
                # users should not override the name for Vertex Pipelines
                # if empty string, replace
                # else provide descriptive warning and prefer letting backend throw exception
                running_on_vertex = 'VERTEX_AI_PIPELINES_RUN_LABELS' in os.environ
                if running_on_vertex:
                    if return_value.name == '':
                        return_value.name = artifact_name
                    else:
                        # prefer letting the backend throw the runtime exception
                        warnings.warn(
                            f'If you are running your pipeline Vertex AI Pipelines, you should not provide a name for your artifact. It will be set to the Vertex artifact resource name {artifact_name} by default. Got value for name: {return_value.name}.',
                            RuntimeWarning,
                            stacklevel=2)
                self.output_artifacts[output_name] = return_value
            else:
                # for -> Artifact annotations, where the user returns some data that the executor should serialize
                self.write_output_artifact_payload(output_name, return_value)
        else:
            raise RuntimeError(
                f'Unknown return type: {annotation_type}. Must be one of the supported data types: https://www.kubeflow.org/docs/components/pipelines/v2/data-types/'
            )

    def write_executor_output(self,
                              func_output: Optional[Any] = None
                             ) -> Optional[str]:
        """Writes executor output containing the Python function output. The
        executor output file will not be written if this code is executed from
        a non-chief node in a mirrored execution strategy.

        Args:
            func_output: The object returned by the function.

        Returns:
            Optional[str]: Returns the location of the executor_output file as a string if the file is written. Else, None.
        """

        if func_output is not None:
            if is_parameter(self.return_annotation) or is_artifact(
                    self.return_annotation):
                # Note: single output is named `Output` in component.yaml.
                self.handle_single_return_value('Output',
                                                self.return_annotation,
                                                func_output)
            elif is_named_tuple(self.return_annotation):
                if len(self.return_annotation._fields) != len(func_output):
                    raise RuntimeError(
                        f'Expected {len(self.return_annotation._fields)} return values from function `{self.func.__name__}`, got {len(func_output)}'
                    )
                for i in range(len(self.return_annotation._fields)):
                    field = self.return_annotation._fields[i]
                    field_type = self.return_annotation.__annotations__[field]
                    if type(func_output) == tuple:
                        field_value = func_output[i]
                    else:
                        field_value = getattr(func_output, field)
                    self.handle_single_return_value(field, field_type,
                                                    field_value)
            else:
                raise RuntimeError(
                    f'Unknown return type: {self.return_annotation}. Must be one of `str`, `int`, `float`, a subclass of `Artifact`, or a NamedTuple collection of these types.'
                )

        if self.output_artifacts:
            self.excutor_output['artifacts'] = {}

        for name, artifact in self.output_artifacts.items():
            runtime_artifact = {
                'name': artifact.name,
                'uri': artifact.uri,
                'metadata': artifact.metadata,
            }
            artifacts_list = {'artifacts': [runtime_artifact]}

            self.excutor_output['artifacts'][name] = artifacts_list

        # This check is to ensure only one worker (in a mirrored, distributed training/compute strategy) attempts to write to the same executor output file at the same time using gcsfuse, which enforces immutability of files.
        write_file = True

        CLUSTER_SPEC_ENV_VAR_NAME = 'CLUSTER_SPEC'
        cluster_spec_string = os.environ.get(CLUSTER_SPEC_ENV_VAR_NAME)
        if cluster_spec_string:
            cluster_spec = json.loads(cluster_spec_string)
            CHIEF_NODE_LABELS = {'workerpool0', 'chief', 'master'}
            write_file = cluster_spec['task']['type'] in CHIEF_NODE_LABELS

        if write_file:
            makedirs_recursively(self.executor_output_path)
            with open(self.executor_output_path, 'w') as f:
                f.write(json.dumps(self.excutor_output))
            return self.executor_output_path
        return None

    def execute(self) -> Optional[str]:
        """Executes the function and writes the executor output file. The
        executor output file will not be written if this code is executed from
        a non-chief node in a mirrored execution strategy.

        Returns:
            Optional[str]: Returns the location of the executor_output file as a string if the file is written. Else, None.
        """
        annotations = inspect.getfullargspec(self.func).annotations

        # Function arguments.
        func_kwargs = {}

        for k, v in annotations.items():
            if k == 'return':
                continue

            # Annotations for parameter types could be written as, for example,
            # `Optional[str]`. In this case, we need to strip off the part
            # `Optional[]` to get the actual parameter type.
            v = type_annotations.maybe_strip_optional_from_annotation(v)
            if v == task_final_status.PipelineTaskFinalStatus:
                value = self.get_input_parameter_value(k)

                # PipelineTaskFinalStatus field names pipelineJobResourceName and pipelineTaskName are deprecated. Support for these fields will be removed at a later date.
                pipline_job_resource_name = 'pipelineJobResourceName'
                if value.get(pipline_job_resource_name) is None:
                    pipline_job_resource_name = 'pipeline_job_resource_name'
                pipeline_task_name = 'pipelineTaskName'
                if value.get(pipeline_task_name) is None:
                    pipeline_task_name = 'pipeline_task_name'

                func_kwargs[k] = task_final_status.PipelineTaskFinalStatus(
                    state=value.get('state'),
                    pipeline_job_resource_name=value.get(
                        pipline_job_resource_name),
                    pipeline_task_name=value.get(pipeline_task_name),
                    error_code=value.get('error', {}).get('code', None),
                    error_message=value.get('error', {}).get('message', None),
                )

            elif type_annotations.is_list_of_artifacts(v):
                func_kwargs[k] = self.get_input_artifact(k)

            elif is_parameter(v):
                value = self.get_input_parameter_value(k)
                if value is not None:
                    func_kwargs[k] = value

            elif type_annotations.is_Input_Output_artifact_annotation(v):
                if type_annotations.is_artifact_wrapped_in_Input(v):
                    func_kwargs[k] = self.get_input_artifact(k)
                if type_annotations.is_artifact_wrapped_in_Output(v):
                    func_kwargs[k] = self.get_output_artifact(k)

            elif is_artifact(v):
                func_kwargs[k] = self.get_input_artifact(k)

            elif isinstance(v, type_annotations.OutputPath):
                if is_parameter(v.type):
                    func_kwargs[k] = self.get_output_parameter_path(k)
                else:
                    func_kwargs[k] = self.get_output_artifact_path(k)

            elif isinstance(v, type_annotations.InputPath):
                func_kwargs[k] = self.get_input_artifact_path(k)

        result = self.func(**func_kwargs)
        return self.write_executor_output(result)


def create_artifact_instance(
    runtime_artifact: Dict,
    fallback_artifact_cls=dsl.Artifact,
) -> type:
    """Creates an artifact class instances from a runtime artifact
    dictionary."""
    schema_title = runtime_artifact.get('type', {}).get('schemaTitle', '')
    artifact_cls = artifact_types._SCHEMA_TITLE_TO_TYPE.get(
        schema_title, fallback_artifact_cls)
    return artifact_cls._from_executor_fields(
        uri=runtime_artifact.get('uri', ''),
        name=runtime_artifact.get('name', ''),
        metadata=runtime_artifact.get('metadata', {}),
    ) if hasattr(artifact_cls, '_from_executor_fields') else artifact_cls(
        uri=runtime_artifact.get('uri', ''),
        name=runtime_artifact.get('name', ''),
        metadata=runtime_artifact.get('metadata', {}),
    )


def get_short_type_name(type_name: str) -> str:
    """Extracts the short form type name.

    This method is used for looking up serializer for a given type.

    For example:
      typing.List -> List
      typing.List[int] -> List
      typing.Dict[str, str] -> Dict
      List -> List
      str -> str

    Args:
      type_name: The original type name.

    Returns:
      The short form type name or the original name if pattern doesn't match.
    """
    match = re.match(r'(typing\.)?(?P<type>\w+)(?:\[.+\])?', type_name)
    return match['type'] if match else type_name


# TODO: merge with type_utils.is_parameter_type
def is_parameter(annotation: Any) -> bool:
    if type(annotation) == type:
        return annotation in [str, int, float, bool, dict, list]

    # Annotation could be, for instance `typing.Dict[str, str]`, etc.
    return get_short_type_name(str(annotation)) in ['Dict', 'List']


def is_artifact(annotation: Any) -> bool:
    if type(annotation) == type:
        return type_annotations.is_artifact_class(annotation)
    return False


def is_named_tuple(annotation: Any) -> bool:
    if type(annotation) == type:
        return issubclass(annotation, tuple) and hasattr(
            annotation, '_fields') and hasattr(annotation, '__annotations__')
    return False


def makedirs_recursively(path: str) -> None:
    os.makedirs(os.path.dirname(path), exist_ok=True)



# Copyright 2023 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for executor_output_utils.py."""

import json
import os
import tempfile
from typing import List
import unittest

from absl.testing import parameterized
from google.protobuf import json_format
from google.protobuf import struct_pb2
from kfp import dsl
from kfp.local import executor_output_utils
from kfp.local import testing_utilities
from kfp.pipeline_spec import pipeline_spec_pb2


class TestGetOutputsFromMessages(
        testing_utilities.LocalRunnerEnvironmentTestCase):

    def test(self):
        executor_input = pipeline_spec_pb2.ExecutorInput()
        json_format.ParseDict(
            {
                'inputs': {
                    'parameterValues': {
                        'string_in': 'foo'
                    }
                },
                'outputs': {
                    'parameters': {
                        'int_out': {
                            'outputFile':
                                'foo/multiple-io-2023-11-09-12-12-05-528112/multiple-io/int_out'
                        },
                        'str_out': {
                            'outputFile':
                                'foo/multiple-io-2023-11-09-12-12-05-528112/multiple-io/str_out'
                        }
                    },
                    'artifacts': {
                        'dataset_out': {
                            'artifacts': [{
                                'name':
                                    'dataset_out',
                                'type': {
                                    'schemaTitle': 'system.Dataset',
                                    'schemaVersion': '0.0.1'
                                },
                                'uri':
                                    'foo/multiple-io-2023-11-09-12-12-05-528112/multiple-io/dataset_out',
                                'metadata': {}
                            }]
                        }
                    },
                    'outputFile':
                        'foo/multiple-io-2023-11-09-12-12-05-528112/multiple-io/executor_output.json'
                }
            }, executor_input)
        component_spec = pipeline_spec_pb2.ComponentSpec()
        json_format.ParseDict(
            {
                'inputDefinitions': {
                    'parameters': {
                        'string_in': {
                            'parameterType': 'STRING'
                        }
                    }
                },
                'outputDefinitions': {
                    'artifacts': {
                        'dataset_out': {
                            'artifactType': {
                                'schemaTitle': 'system.Dataset',
                                'schemaVersion': '0.0.1'
                            }
                        }
                    },
                    'parameters': {
                        'int_out': {
                            'parameterType': 'NUMBER_INTEGER'
                        },
                        'str_out': {
                            'parameterType': 'STRING'
                        }
                    }
                },
                'executorLabel': 'exec-multiple-io'
            }, component_spec)
        executor_output = pipeline_spec_pb2.ExecutorOutput()
        json_format.ParseDict(
            {
                'parameterValues': {
                    'int_out': 1,
                    'str_out': 'foo'
                },
                'artifacts': {
                    'dataset_out': {
                        'artifacts': [{
                            'name':
                                'dataset_out',
                            'uri':
                                'foo/multiple-io-2023-11-09-12-12-05-528112/multiple-io/dataset_out',
                            'metadata': {
                                'foo': 'bar'
                            }
                        }]
                    }
                }
            }, executor_output)

        os.makedirs(os.path.dirname(executor_input.outputs.output_file))
        with open(executor_input.outputs.output_file, 'w') as f:
            f.write(json_format.MessageToJson(executor_output))
        outputs = executor_output_utils.get_outputs_for_task(
            executor_input=executor_input,
            component_spec=component_spec,
        )

        self.assertEqual(outputs['int_out'], 1)
        self.assertEqual(outputs['str_out'], 'foo')
        assert_artifacts_equal(
            self,
            outputs['dataset_out'],
            dsl.Dataset(
                name='dataset_out',
                uri='foo/multiple-io-2023-11-09-12-12-05-528112/multiple-io/dataset_out',
                metadata={'foo': 'bar'}),
        )


class TestLoadExecutorOutput(unittest.TestCase):

    def test_exists(self):
        with tempfile.TemporaryDirectory() as tempdir:
            executor_output = pipeline_spec_pb2.ExecutorOutput(
                parameter_values={
                    'foo': struct_pb2.Value(string_value='foo_value')
                })
            path = os.path.join(tempdir, 'executor_output.json')
            testing_utilities.write_proto_to_json_file(executor_output, path)

            actual = executor_output_utils.load_executor_output(path)
            expected = pipeline_spec_pb2.ExecutorOutput()
            expected.parameter_values['foo'].CopyFrom(
                struct_pb2.Value(string_value='foo_value'))
            self.assertEqual(
                actual.SerializeToString(deterministic=True),
                expected.SerializeToString(deterministic=True),
            )

    def test_not_exists(self):
        non_existent_path = 'non_existent_path.json'
        actual = executor_output_utils.load_executor_output(non_existent_path)
        expected = pipeline_spec_pb2.ExecutorOutput()
        self.assertEqual(
            actual.SerializeToString(deterministic=True),
            expected.SerializeToString(deterministic=True),
        )


class TestGetOutputsFromExecutorOutput(unittest.TestCase):

    def test_param_and_artifact_outputs(self):
        # include the special case of an output int for more complete testing of behavior
        executor_output = pipeline_spec_pb2.ExecutorOutput()
        json_format.ParseDict(
            {
                'parameterValues': {
                    'int_out': 1,
                    'str_out': 'foo'
                },
                'artifacts': {
                    'dataset_out': {
                        'artifacts': [{
                            'name':
                                'dataset_out',
                            'uri':
                                'foo/multiple-io-2023-11-09-11-31-31-064429/multiple-io/dataset_out',
                            'metadata': {
                                'foo': 'bar'
                            }
                        }]
                    }
                }
            }, executor_output)
        executor_input = pipeline_spec_pb2.ExecutorInput()
        json_format.ParseDict(
            {
                'inputs': {
                    'parameterValues': {
                        'string_in': 'foo'
                    }
                },
                'outputs': {
                    'parameters': {
                        'int_out': {
                            'outputFile':
                                'foo/temp_root/multiple-io-2023-11-09-11-31-31-064429/multiple-io/int_out'
                        },
                        'str_out': {
                            'outputFile':
                                'foo/multiple-io-2023-11-09-11-31-31-064429/multiple-io/str_out'
                        }
                    },
                    'artifacts': {
                        'dataset_out': {
                            'artifacts': [{
                                'name':
                                    'dataset_out',
                                'type': {
                                    'schemaTitle': 'system.Dataset',
                                    'schemaVersion': '0.0.1'
                                },
                                'uri':
                                    'foo/multiple-io-2023-11-09-11-31-31-064429/multiple-io/dataset_out',
                                'metadata': {}
                            }]
                        }
                    },
                    'outputFile':
                        'foo/multiple-io-2023-11-09-11-31-31-064429/multiple-io/executor_output.json'
                }
            }, executor_input)
        component_spec = pipeline_spec_pb2.ComponentSpec()
        json_format.ParseDict(
            {
                'inputDefinitions': {
                    'parameters': {
                        'string_in': {
                            'parameterType': 'STRING'
                        }
                    }
                },
                'outputDefinitions': {
                    'artifacts': {
                        'dataset_out': {
                            'artifactType': {
                                'schemaTitle': 'system.Dataset',
                                'schemaVersion': '0.0.1'
                            }
                        }
                    },
                    'parameters': {
                        'int_out': {
                            'parameterType': 'NUMBER_INTEGER'
                        },
                        'str_out': {
                            'parameterType': 'STRING'
                        }
                    }
                },
                'executorLabel': 'exec-multiple-io'
            }, component_spec)

        outputs = executor_output_utils.get_outputs_from_executor_output(
            executor_output=executor_output,
            executor_input=executor_input,
            component_spec=component_spec,
        )
        self.assertIsInstance(outputs, dict)
        self.assertIsInstance(outputs['dataset_out'], dsl.Dataset)
        self.assertEqual(outputs['dataset_out'].name, 'dataset_out')
        self.assertEqual(
            outputs['dataset_out'].uri,
            'foo/multiple-io-2023-11-09-11-31-31-064429/multiple-io/dataset_out'
        )
        self.assertEqual(outputs['dataset_out'].metadata, {'foo': 'bar'})
        self.assertEqual(outputs['int_out'], 1)
        self.assertEqual(outputs['str_out'], 'foo')


class TestPb2ValueToPython(unittest.TestCase):

    def test_null(self):
        inp = struct_pb2.Value(null_value=struct_pb2.NullValue.NULL_VALUE)
        actual = executor_output_utils.pb2_value_to_python(inp)
        expected = None
        self.assertEqual(actual, expected)

    def test_string(self):
        inp = struct_pb2.Value(string_value='foo_value')
        actual = executor_output_utils.pb2_value_to_python(inp)
        expected = 'foo_value'
        self.assertEqual(actual, expected)

    def test_number_int(self):
        inp = struct_pb2.Value(number_value=1)
        actual = executor_output_utils.pb2_value_to_python(inp)
        expected = 1.0
        self.assertEqual(actual, expected)

    def test_number_float(self):
        inp = struct_pb2.Value(number_value=1.0)
        actual = executor_output_utils.pb2_value_to_python(inp)
        expected = 1.0
        self.assertEqual(actual, expected)

    def test_bool(self):
        inp = struct_pb2.Value(bool_value=True)
        actual = executor_output_utils.pb2_value_to_python(inp)
        expected = True
        self.assertIs(actual, expected)

    def test_dict(self):
        struct_value = struct_pb2.Struct()
        struct_value.fields['my_key'].string_value = 'my_value'
        struct_value.fields['other_key'].bool_value = True
        inp = struct_pb2.Value(struct_value=struct_value)
        actual = executor_output_utils.pb2_value_to_python(inp)
        expected = {'my_key': 'my_value', 'other_key': True}
        self.assertEqual(actual, expected)


class TestRuntimeArtifactToDslArtifact(unittest.TestCase):

    def test_artifact(self):
        metadata = struct_pb2.Struct()
        metadata.fields['foo'].string_value = 'bar'
        type_ = pipeline_spec_pb2.ArtifactTypeSchema(
            schema_title='system.Artifact',
            schema_version='0.0.1',
        )
        runtime_artifact = pipeline_spec_pb2.RuntimeArtifact(
            name='a',
            uri='gs://bucket/foo',
            metadata=metadata,
            type=type_,
        )
        actual = executor_output_utils.runtime_artifact_to_dsl_artifact(
            runtime_artifact)
        expected = dsl.Artifact(
            name='a',
            uri='gs://bucket/foo',
            metadata={'foo': 'bar'},
        )
        assert_artifacts_equal(self, actual, expected)

    def test_dataset(self):
        metadata = struct_pb2.Struct()
        metadata.fields['baz'].string_value = 'bat'
        type_ = pipeline_spec_pb2.ArtifactTypeSchema(
            schema_title='system.Dataset',
            schema_version='0.0.1',
        )
        runtime_artifact = pipeline_spec_pb2.RuntimeArtifact(
            name='d',
            uri='gs://bucket/foo',
            metadata=metadata,
            type=type_,
        )
        actual = executor_output_utils.runtime_artifact_to_dsl_artifact(
            runtime_artifact)
        expected = dsl.Dataset(
            name='d',
            uri='gs://bucket/foo',
            metadata={'baz': 'bat'},
        )
        assert_artifacts_equal(self, actual, expected)


class TestArtifactListToDslArtifact(unittest.TestCase):

    def test_not_list(self):
        metadata = struct_pb2.Struct()
        metadata.fields['foo'].string_value = 'bar'
        type_ = pipeline_spec_pb2.ArtifactTypeSchema(
            schema_title='system.Artifact',
            schema_version='0.0.1',
        )
        runtime_artifact = pipeline_spec_pb2.RuntimeArtifact(
            name='a',
            uri='gs://bucket/foo',
            metadata=metadata,
            type=type_,
        )
        artifact_list = pipeline_spec_pb2.ArtifactList(
            artifacts=[runtime_artifact])

        actual = executor_output_utils.artifact_list_to_dsl_artifact(
            artifact_list,
            is_artifact_list=False,
        )
        expected = dsl.Artifact(
            name='a',
            uri='gs://bucket/foo',
            metadata={'foo': 'bar'},
        )
        assert_artifacts_equal(self, actual, expected)

    def test_single_entry_list(self):
        metadata = struct_pb2.Struct()
        metadata.fields['foo'].string_value = 'bar'
        type_ = pipeline_spec_pb2.ArtifactTypeSchema(
            schema_title='system.Dataset',
            schema_version='0.0.1',
        )
        runtime_artifact = pipeline_spec_pb2.RuntimeArtifact(
            name='a',
            uri='gs://bucket/foo',
            metadata=metadata,
            type=type_,
        )
        artifact_list = pipeline_spec_pb2.ArtifactList(
            artifacts=[runtime_artifact])

        actual = executor_output_utils.artifact_list_to_dsl_artifact(
            artifact_list,
            is_artifact_list=True,
        )
        expected = [
            dsl.Dataset(
                name='a',
                uri='gs://bucket/foo',
                metadata={'foo': 'bar'},
            )
        ]
        assert_artifact_lists_equal(self, actual, expected)

    def test_multi_entry_list(self):
        metadata = struct_pb2.Struct()
        metadata.fields['foo'].string_value = 'bar'
        type_ = pipeline_spec_pb2.ArtifactTypeSchema(
            schema_title='system.Dataset',
            schema_version='0.0.1',
        )
        runtime_artifact1 = pipeline_spec_pb2.RuntimeArtifact(
            name='a',
            uri='gs://bucket/foo/a',
            metadata=metadata,
            type=type_,
        )
        runtime_artifact2 = pipeline_spec_pb2.RuntimeArtifact(
            name='b',
            uri='gs://bucket/foo/b',
            type=type_,
        )
        artifact_list = pipeline_spec_pb2.ArtifactList(
            artifacts=[runtime_artifact1, runtime_artifact2])

        actual = executor_output_utils.artifact_list_to_dsl_artifact(
            artifact_list,
            is_artifact_list=True,
        )
        expected = [
            dsl.Dataset(
                name='a',
                uri='gs://bucket/foo/a',
                metadata={'foo': 'bar'},
            ),
            dsl.Dataset(
                name='b',
                uri='gs://bucket/foo/b',
            )
        ]

        assert_artifact_lists_equal(self, actual, expected)


class AddTypeToExecutorOutput(unittest.TestCase):

    def test(self):
        executor_input = pipeline_spec_pb2.ExecutorInput()
        json_format.ParseDict(
            {
                'inputs': {},
                'outputs': {
                    'artifacts': {
                        'dataset_out': {
                            'artifacts': [{
                                'name':
                                    'dataset_out',
                                'type': {
                                    'schemaTitle': 'system.Dataset',
                                    'schemaVersion': '0.0.1'
                                },
                                'uri':
                                    'foo/multiple-io-2023-11-09-12-04-18-616263/multiple-io/dataset_out',
                                'metadata': {}
                            }]
                        },
                        'model_out': {
                            'artifacts': [{
                                'name':
                                    'model_out',
                                'type': {
                                    'schemaTitle': 'system.Model',
                                    'schemaVersion': '0.0.1'
                                },
                                'uri':
                                    'foo/multiple-io-2023-11-09-12-04-18-616263/multiple-io/model_out',
                                'metadata': {}
                            }]
                        }
                    },
                    'outputFile':
                        'foo/multiple-io-2023-11-09-12-04-18-616263/multiple-io/executor_output.json'
                }
            }, executor_input)
        executor_output = pipeline_spec_pb2.ExecutorOutput()
        json_format.ParseDict(
            {
                'artifacts': {
                    'dataset_out': {
                        'artifacts': [{
                            'name':
                                'dataset_out',
                            'uri':
                                'foo/multiple-io-2023-11-09-12-04-18-616263/multiple-io/dataset_out',
                            'metadata': {
                                'foo': 'bar'
                            }
                        }]
                    },
                    'model_out': {
                        'artifacts': [{
                            'name':
                                'model_out',
                            'uri':
                                'foo/multiple-io-2023-11-09-12-04-18-616263/multiple-io/model_out',
                            'metadata': {
                                'baz': 'bat'
                            }
                        }]
                    }
                }
            }, executor_output)

        expected = pipeline_spec_pb2.ExecutorOutput()
        json_format.ParseDict(
            {
                'artifacts': {
                    'dataset_out': {
                        'artifacts': [{
                            'name':
                                'dataset_out',
                            'uri':
                                'foo/multiple-io-2023-11-09-12-04-18-616263/multiple-io/dataset_out',
                            'metadata': {
                                'foo': 'bar'
                            },
                            'type': {
                                'schemaTitle': 'system.Dataset',
                                'schemaVersion': '0.0.1'
                            },
                        }]
                    },
                    'model_out': {
                        'artifacts': [{
                            'name':
                                'model_out',
                            'uri':
                                'foo/multiple-io-2023-11-09-12-04-18-616263/multiple-io/model_out',
                            'metadata': {
                                'baz': 'bat'
                            },
                            'type': {
                                'schemaTitle': 'system.Model',
                                'schemaVersion': '0.0.1'
                            },
                        }]
                    }
                }
            }, expected)

        actual = executor_output_utils.add_type_to_executor_output(
            executor_input=executor_input,
            executor_output=executor_output,
        )
        self.assertEqual(actual, expected)


class TestSpecialDslOutputPathRead(parameterized.TestCase):

    @parameterized.parameters([
        ('foo', 'foo',
         pipeline_spec_pb2.ParameterType.ParameterTypeEnum.STRING),
        ('foo', 'foo',
         pipeline_spec_pb2.ParameterType.ParameterTypeEnum.STRING),
        ('true', True,
         pipeline_spec_pb2.ParameterType.ParameterTypeEnum.BOOLEAN),
        ('True', True,
         pipeline_spec_pb2.ParameterType.ParameterTypeEnum.BOOLEAN),
        ('false', False,
         pipeline_spec_pb2.ParameterType.ParameterTypeEnum.BOOLEAN),
        ('False', False,
         pipeline_spec_pb2.ParameterType.ParameterTypeEnum.BOOLEAN),
        (json.dumps({'x': 'y'}), {
            'x': 'y'
        }, pipeline_spec_pb2.ParameterType.ParameterTypeEnum.STRUCT),
        ('3.14', 3.14,
         pipeline_spec_pb2.ParameterType.ParameterTypeEnum.NUMBER_DOUBLE),
        ('100', 100,
         pipeline_spec_pb2.ParameterType.ParameterTypeEnum.NUMBER_INTEGER),
    ])
    def test(self, written, expected, dtype):
        with tempfile.TemporaryDirectory() as tempdir:
            output_file = os.path.join(tempdir, 'Output')
            with open(output_file, 'w') as f:
                f.write(written)

            actual = executor_output_utils.special_dsl_outputpath_read(
                parameter_name='name',
                output_file=output_file,
                dtype=dtype,
            )

        self.assertEqual(actual, expected)

    def test_exception(self):
        with tempfile.TemporaryDirectory() as tempdir:
            output_file = os.path.join(tempdir, 'Output')
            with open(output_file, 'w') as f:
                f.write(str({'x': 'y'}))
            with self.assertRaisesRegex(
                    ValueError,
                    r"Could not deserialize output 'name' from path"):
                executor_output_utils.special_dsl_outputpath_read(
                    parameter_name='name',
                    output_file=output_file,
                    dtype=pipeline_spec_pb2.ParameterType.ParameterTypeEnum
                    .STRUCT,
                )


def assert_artifacts_equal(
    test_class: unittest.TestCase,
    a1: dsl.Artifact,
    a2: dsl.Artifact,
) -> None:
    test_class.assertEqual(a1.name, a2.name)
    test_class.assertEqual(a1.uri, a2.uri)
    test_class.assertEqual(a1.metadata, a2.metadata)
    test_class.assertEqual(a1.schema_title, a2.schema_title)
    test_class.assertEqual(a1.schema_version, a2.schema_version)
    test_class.assertIsInstance(a1, type(a2))


def assert_artifact_lists_equal(
    test_class: unittest.TestCase,
    l1: List[dsl.Artifact],
    l2: List[dsl.Artifact],
) -> None:
    test_class.assertEqual(len(l1), len(l2))
    for a1, a2 in zip(l1, l2):
        assert_artifacts_equal(test_class, a1, a2)


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Importer-based component."""

from kfp.dsl import base_component
from kfp.dsl import structures


class ImporterComponent(base_component.BaseComponent):
    """Component defined via dsl.importer."""

    def __init__(
        self,
        component_spec: structures.ComponentSpec,
    ):
        super().__init__(component_spec=component_spec)

    def execute(self, **kwargs):
        raise NotImplementedError



# Copyright 2020-2022 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

warnings.warn(
    (f'The module `{__name__}` is deprecated and will be removed in a future'
     'version. Please import directly from the `kfp` namespace, '
     'instead of `kfp.v2`.'),
    category=DeprecationWarning,
    stacklevel=2)

from kfp import compiler  # noqa: keep unused import
from kfp import components  # noqa: keep unused import
from kfp import dsl  # noqa: keep unused import



# Copyright 2021-2022 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
import inspect
import itertools
import pathlib
import re
import textwrap
from typing import (Any, Callable, Dict, List, Mapping, Optional, Tuple, Type,
                    Union)
import warnings

import docstring_parser
import kfp
from kfp import dsl
from kfp.dsl import container_component_artifact_channel
from kfp.dsl import container_component_class
from kfp.dsl import graph_component
from kfp.dsl import pipeline_config
from kfp.dsl import placeholders
from kfp.dsl import python_component
from kfp.dsl import structures
from kfp.dsl import task_final_status
from kfp.dsl.types import artifact_types
from kfp.dsl.types import custom_artifact_types
from kfp.dsl.types import type_annotations
from kfp.dsl.types import type_utils

_DEFAULT_BASE_IMAGE = 'python:3.9'
SINGLE_OUTPUT_NAME = 'Output'


@dataclasses.dataclass
class ComponentInfo():
    """A dataclass capturing registered components.

    This will likely be subsumed/augmented with BaseComponent.
    """
    name: str
    function_name: str
    func: Callable
    target_image: str
    module_path: pathlib.Path
    component_spec: structures.ComponentSpec
    output_component_file: Optional[str] = None
    base_image: str = _DEFAULT_BASE_IMAGE
    packages_to_install: Optional[List[str]] = None
    pip_index_urls: Optional[List[str]] = None
    pip_trusted_hosts: Optional[List[str]] = None
    use_venv: bool = False


# A map from function_name to components.  This is always populated when a
# module containing KFP components is loaded. Primarily used by KFP CLI
# component builder to package components in a file into containers.
REGISTERED_MODULES = None


def _python_function_name_to_component_name(name):
    name_with_spaces = re.sub(' +', ' ', name.replace('_', ' ')).strip(' ')
    return name_with_spaces[0].upper() + name_with_spaces[1:]


def make_index_url_options(pip_index_urls: Optional[List[str]],
                           pip_trusted_hosts: Optional[List[str]]) -> str:
    """Generates index URL options for the pip install command based on the
    provided pip_index_urls and pip_trusted_hosts.

    Args:
        pip_index_urls (Optional[List[str]]): Optional list of pip index URLs.
        pip_trusted_hosts (Optional[List[str]]): Optional list of pip trusted hosts.

    Returns:
        str:
            - An empty string if pip_index_urls is empty or None.
            - '--index-url url ' if pip_index_urls contains 1 URL.
            - The above followed by '--extra-index-url url ' for each additional URL in pip_index_urls
            if pip_index_urls contains more than 1 URL.
            - If pip_trusted_hosts is None:
                - The above followed by '--trusted-host url ' for each URL in pip_index_urls.
            - If pip_trusted_hosts is an empty List.
                - No --trusted-host information will be added
            - If pip_trusted_hosts contains any URLs:
                - The above followed by '--trusted-host url ' for each URL in pip_trusted_hosts.
    Note:
        In case pip_index_urls is not empty, the returned string will contain a space at the end.
    """
    if not pip_index_urls:
        return ''

    index_url = pip_index_urls[0]
    extra_index_urls = pip_index_urls[1:]

    options = [f'--index-url {index_url}']
    options.extend(f'--extra-index-url {extra_index_url}'
                   for extra_index_url in extra_index_urls)

    if pip_trusted_hosts is None:
        options.extend([f'--trusted-host {index_url}'])
        options.extend(f'--trusted-host {extra_index_url}'
                       for extra_index_url in extra_index_urls)
    elif len(pip_trusted_hosts) > 0:
        options.extend(f'--trusted-host {trusted_host}'
                       for trusted_host in pip_trusted_hosts)

    return ' '.join(options) + ' '


def make_pip_install_command(
    install_parts: List[str],
    index_url_options: str,
) -> str:
    concat_package_list = ' '.join(
        [repr(str(package)) for package in install_parts])
    return f'python3 -m pip install --quiet --no-warn-script-location {index_url_options}{concat_package_list}'


_install_python_packages_script_template = '''
if ! [ -x "$(command -v pip)" ]; then
    python3 -m ensurepip || python3 -m ensurepip --user || apt-get install python3-pip
fi

PIP_DISABLE_PIP_VERSION_CHECK=1 {pip_install_commands} && "$0" "$@"
'''

# Creates and activates a virtual environment in a temporary directory.
# The environment inherits the system site packages.
_use_venv_script_template = '''
export PIP_DISABLE_PIP_VERSION_CHECK=1
tmp=$(mktemp -d)
python3 -m venv "$tmp/venv" --system-site-packages
. "$tmp/venv/bin/activate"
'''


def _get_packages_to_install_command(
    kfp_package_path: Optional[str] = None,
    pip_index_urls: Optional[List[str]] = None,
    packages_to_install: Optional[List[str]] = None,
    install_kfp_package: bool = True,
    target_image: Optional[str] = None,
    pip_trusted_hosts: Optional[List[str]] = None,
    use_venv: bool = False,
) -> List[str]:
    packages_to_install = packages_to_install or []
    kfp_in_user_pkgs = any(pkg.startswith('kfp') for pkg in packages_to_install)
    # if the user doesn't say "don't install", they aren't building a
    # container component, and they haven't already specified a KFP dep
    # themselves, we install KFP for them
    inject_kfp_install = install_kfp_package and target_image is None and not kfp_in_user_pkgs
    if not inject_kfp_install and not packages_to_install:
        return []
    pip_install_strings = []
    index_url_options = make_index_url_options(pip_index_urls,
                                               pip_trusted_hosts)

    # Install packages before KFP. This allows us to
    # control where we source kfp-pipeline-spec.
    # This is particularly useful for development and
    # CI use-case when you want to install the spec
    # from source.
    if packages_to_install:
        user_packages_pip_install_command = make_pip_install_command(
            install_parts=packages_to_install,
            index_url_options=index_url_options,
        )
        pip_install_strings.append(user_packages_pip_install_command)
        if inject_kfp_install:
            pip_install_strings.append(' && ')

    if inject_kfp_install:
        if use_venv:
            pip_install_strings.append(_use_venv_script_template)
        if kfp_package_path:
            kfp_pip_install_command = make_pip_install_command(
                install_parts=[kfp_package_path],
                index_url_options=index_url_options,
            )
        else:
            kfp_pip_install_command = make_pip_install_command(
                install_parts=[
                    f'kfp=={kfp.__version__}',
                    '--no-deps',
                    'typing-extensions>=3.7.4,<5; python_version<"3.9"',
                ],
                index_url_options=index_url_options,
            )
        pip_install_strings.append(kfp_pip_install_command)

    return [
        'sh', '-c',
        _install_python_packages_script_template.format(
            pip_install_commands=' '.join(pip_install_strings))
    ]


def _get_function_source_definition(func: Callable) -> str:
    func_code = inspect.getsource(func)

    # Function might be defined in some indented scope (e.g. in another
    # function). We need to handle this and properly dedent the function source
    # code
    func_code = textwrap.dedent(func_code)
    func_code_lines = func_code.split('\n')

    # Removing possible decorators (can be multiline) until the function
    # definition is found
    func_code_lines = itertools.dropwhile(lambda x: not x.startswith('def'),
                                          func_code_lines)

    if not func_code_lines:
        raise ValueError(
            f'Failed to dedent and clean up the source of function "{func.__name__}". It is probably not properly indented.'
        )

    return '\n'.join(func_code_lines)


def maybe_make_unique(name: str, names: List[str]):
    if name not in names:
        return name

    for i in range(2, 100):
        unique_name = f'{name}_{i}'
        if unique_name not in names:
            return unique_name

    raise RuntimeError(f'Too many arguments with the name {name}')


def get_name_to_specs(
    signature: inspect.Signature,
    containerized: bool = False,
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """Returns two dictionaries.

    The first is a mapping of input name to input annotation. The second
    is a mapping of output name to output annotation.
    """
    func_params = list(signature.parameters.values())

    name_to_input_specs = {}
    name_to_output_specs = {}

    ### handle function parameter annotations ###
    for func_param in func_params:
        name = func_param.name
        if name == SINGLE_OUTPUT_NAME:
            raise ValueError(
                f'"{SINGLE_OUTPUT_NAME}" is an invalid parameter name.')
        # Stripping Optional from Optional[<type>] is the only processing done
        # on annotations in this flow. Other than that, we extract the raw
        # annotation and process later.
        annotation = type_annotations.maybe_strip_optional_from_annotation(
            func_param.annotation)

        # no annotation
        if annotation == inspect._empty:
            raise TypeError(f'Missing type annotation for argument: {name}')

        # is Input[Artifact], Input[List[<Artifact>]], <param> (e.g., str), or InputPath(<param>)
        elif (type_annotations.is_artifact_wrapped_in_Input(annotation) or
              isinstance(
                  annotation,
                  type_annotations.InputPath,
              ) or type_utils.is_parameter_type(annotation)):
            name_to_input_specs[maybe_make_unique(
                name, list(name_to_input_specs))] = make_input_spec(
                    annotation, func_param)
        # is Artifact annotation (e.g., Artifact, Dataset, etc.)
        # or List[<Artifact>]
        elif type_annotations.issubclass_of_artifact(
                annotation) or type_annotations.is_list_of_artifacts(
                    annotation):
            if containerized:
                raise TypeError(
                    f"Container Components must wrap input and output artifact annotations with Input/Output type markers (Input[<artifact>] or Output[<artifact>]). Got function input '{name}' with annotation {annotation}."
                )
            name_to_input_specs[maybe_make_unique(
                name, list(name_to_input_specs))] = make_input_spec(
                    annotation, func_param)

        # is Output[Artifact] or OutputPath(<param>)
        elif type_annotations.is_artifact_wrapped_in_Output(
                annotation) or isinstance(annotation,
                                          type_annotations.OutputPath):
            name_to_output_specs[maybe_make_unique(
                name,
                list(name_to_output_specs))] = make_output_spec(annotation)

        # parameter type
        else:
            type_string = type_utils._annotation_to_type_struct(annotation)
            name_to_input_specs[maybe_make_unique(
                name, list(name_to_input_specs))] = make_input_spec(
                    type_string, func_param)

    ### handle return annotations ###
    return_ann = signature.return_annotation

    # validate container component returns
    if containerized:
        if return_ann not in [
                inspect.Parameter.empty,
                structures.ContainerSpec,
        ]:
            raise TypeError(
                'Return annotation should be either ContainerSpec or omitted for container components.'
            )
    # ignore omitted returns
    elif return_ann is None or return_ann == inspect.Parameter.empty:
        pass
    # is NamedTuple
    elif hasattr(return_ann, '_fields'):
        # Getting field type annotations.
        # __annotations__ does not exist in python 3.5 and earlier
        # _field_types does not exist in python 3.9 and later
        field_annotations = getattr(return_ann, '__annotations__',
                                    None) or getattr(return_ann, '_field_types')
        for name in return_ann._fields:
            annotation = field_annotations[name]
            if not type_annotations.is_list_of_artifacts(
                    annotation) and not type_annotations.is_artifact_class(
                        annotation):
                annotation = type_utils._annotation_to_type_struct(annotation)
            name_to_output_specs[maybe_make_unique(
                name,
                list(name_to_output_specs))] = make_output_spec(annotation)
    # is deprecated dict returns style
    elif isinstance(return_ann, dict):
        warnings.warn(
            'The ability to specify multiple outputs using the dict syntax'
            ' has been deprecated. It will be removed soon after release'
            ' 0.1.32. Please use typing.NamedTuple to declare multiple'
            ' outputs.', DeprecationWarning)
        for output_name, output_type_annotation in return_ann.items():
            output_type = type_utils._annotation_to_type_struct(
                output_type_annotation)
            name_to_output_specs[maybe_make_unique(
                output_name, list(name_to_output_specs))] = output_type
    # is the simple single return case (can be `-> <param>` or `-> Artifact`)
    # treated the same way, since processing is done in inner functions
    else:
        name_to_output_specs[maybe_make_unique(
            SINGLE_OUTPUT_NAME,
            list(name_to_output_specs))] = make_output_spec(return_ann)
    return name_to_input_specs, name_to_output_specs


def canonicalize_annotation(annotation: Any):
    """Does cleaning on annotations that are common between input and output
    annotations."""
    if type_annotations.is_Input_Output_artifact_annotation(annotation):
        annotation = type_annotations.strip_Input_or_Output_marker(annotation)
    if isinstance(annotation,
                  (type_annotations.InputPath, type_annotations.OutputPath)):
        annotation = annotation.type
    return annotation


def make_input_output_spec_args(annotation: Any) -> Dict[str, Any]:
    """Gets a dict of kwargs shared between InputSpec and OutputSpec."""
    is_artifact_list = type_annotations.is_list_of_artifacts(annotation)
    if is_artifact_list:
        annotation = type_annotations.get_inner_type(annotation)

    if type_annotations.issubclass_of_artifact(annotation):
        typ = type_utils.create_bundled_artifact_type(annotation.schema_title,
                                                      annotation.schema_version)
    else:
        typ = type_utils._annotation_to_type_struct(annotation)
    return {'type': typ, 'is_artifact_list': is_artifact_list}


def make_output_spec(annotation: Any) -> structures.OutputSpec:
    annotation = canonicalize_annotation(annotation)
    args = make_input_output_spec_args(annotation)
    return structures.OutputSpec(**args)


def make_input_spec(annotation: Any,
                    inspect_param: inspect.Parameter) -> structures.InputSpec:
    """Makes an InputSpec from a cleaned output annotation."""
    annotation = canonicalize_annotation(annotation)
    input_output_spec_args = make_input_output_spec_args(annotation)

    if (type_annotations.issubclass_of_artifact(annotation) or
            input_output_spec_args['is_artifact_list']
       ) and inspect_param.default not in {None, inspect._empty}:
        raise ValueError(
            f'Optional Input artifacts may only have default value None. Got: {inspect_param.default}.'
        )

    default = None if inspect_param.default == inspect.Parameter.empty or type_annotations.issubclass_of_artifact(
        annotation) else inspect_param.default

    optional = inspect_param.default is not inspect.Parameter.empty or type_utils.is_task_final_status_type(
        getattr(inspect_param.annotation, '__name__', ''))
    return structures.InputSpec(
        **input_output_spec_args,
        default=default,
        optional=optional,
    )


def extract_component_interface(
    func: Callable,
    containerized: bool = False,
    description: Optional[str] = None,
    name: Optional[str] = None,
) -> structures.ComponentSpec:

    def assign_descriptions(
        inputs_or_outputs: Mapping[str, Union[structures.InputSpec,
                                              structures.OutputSpec]],
        docstring_params: List[docstring_parser.DocstringParam],
    ) -> None:
        """Assigns descriptions to InputSpec or OutputSpec for each component
        input/output found in the parsed docstring parameters."""
        docstring_inputs = {param.arg_name: param for param in docstring_params}
        for name, spec in inputs_or_outputs.items():
            if name in docstring_inputs:
                spec.description = docstring_inputs[name].description

    def parse_docstring_with_return_as_args(
            docstring: Union[str,
                             None]) -> Optional[docstring_parser.Docstring]:
        """Modifies docstring so that a return section can be treated as an
        args section, then parses the docstring."""
        if docstring is None:
            return None

        # Returns and Return are the only two keywords docstring_parser uses for returns
        # use newline to avoid replacements that aren't in the return section header
        return_keywords = ['Returns:\n', 'Returns\n', 'Return:\n', 'Return\n']
        for keyword in return_keywords:
            if keyword in docstring:
                modified_docstring = docstring.replace(keyword.strip(), 'Args:')
                return docstring_parser.parse(modified_docstring)

        return None

    signature = inspect.signature(func)
    name_to_input_spec, name_to_output_spec = get_name_to_specs(
        signature, containerized)
    original_docstring = inspect.getdoc(func)
    parsed_docstring = docstring_parser.parse(original_docstring)

    assign_descriptions(name_to_input_spec, parsed_docstring.params)

    modified_parsed_docstring = parse_docstring_with_return_as_args(
        original_docstring)
    if modified_parsed_docstring is not None:
        assign_descriptions(name_to_output_spec,
                            modified_parsed_docstring.params)

    description = get_pipeline_description(
        decorator_description=description,
        docstring=parsed_docstring,
    )

    component_name = name or _python_function_name_to_component_name(
        func.__name__)
    return structures.ComponentSpec(
        name=component_name,
        description=description,
        inputs=name_to_input_spec or None,
        outputs=name_to_output_spec or None,
        implementation=structures.Implementation(),
    )


EXECUTOR_MODULE = 'kfp.dsl.executor_main'
CONTAINERIZED_PYTHON_COMPONENT_COMMAND = [
    'python3',
    '-m',
    EXECUTOR_MODULE,
]


def _get_command_and_args_for_lightweight_component(
        func: Callable) -> Tuple[List[str], List[str]]:
    imports_source = [
        'import kfp',
        'from kfp import dsl',
        'from kfp.dsl import *',
        'from typing import *',
    ] + custom_artifact_types.get_custom_artifact_type_import_statements(func)

    func_source = _get_function_source_definition(func)
    source = textwrap.dedent('''
        {imports_source}

        {func_source}\n''').format(
        imports_source='\n'.join(imports_source), func_source=func_source)
    command = [
        'sh',
        '-ec',
        textwrap.dedent(f'''\
                    program_path=$(mktemp -d)

                    printf "%s" "$0" > "$program_path/ephemeral_component.py"
                    _KFP_RUNTIME=true python3 -m {EXECUTOR_MODULE} \
                        --component_module_path \
                        "$program_path/ephemeral_component.py" \
                        "$@"
                '''),
        source,
    ]

    args = [
        '--executor_input',
        dsl.PIPELINE_TASK_EXECUTOR_INPUT_PLACEHOLDER,
        '--function_to_execute',
        func.__name__,
    ]

    return command, args


def _get_command_and_args_for_containerized_component(
        function_name: str) -> Tuple[List[str], List[str]]:

    args = [
        '--executor_input',
        dsl.PIPELINE_TASK_EXECUTOR_INPUT_PLACEHOLDER,
        '--function_to_execute',
        function_name,
    ]
    return CONTAINERIZED_PYTHON_COMPONENT_COMMAND, args


def create_component_from_func(
    func: Callable,
    base_image: Optional[str] = None,
    target_image: Optional[str] = None,
    packages_to_install: List[str] = None,
    pip_index_urls: Optional[List[str]] = None,
    output_component_file: Optional[str] = None,
    install_kfp_package: bool = True,
    kfp_package_path: Optional[str] = None,
    pip_trusted_hosts: Optional[List[str]] = None,
    use_venv: bool = False,
) -> python_component.PythonComponent:
    """Implementation for the @component decorator.

    The decorator is defined under component_decorator.py. See the
    decorator for the canonical documentation for this function.
    """

    packages_to_install_command = _get_packages_to_install_command(
        install_kfp_package=install_kfp_package,
        target_image=target_image,
        kfp_package_path=kfp_package_path,
        packages_to_install=packages_to_install,
        pip_index_urls=pip_index_urls,
        pip_trusted_hosts=pip_trusted_hosts,
        use_venv=use_venv,
    )

    command = []
    args = []
    if base_image is None:
        base_image = _DEFAULT_BASE_IMAGE
        warnings.warn(
            ("The default base_image used by the @dsl.component decorator will switch from 'python:3.9' to 'python:3.10' on Oct 1, 2025. To ensure your existing components work with versions of the KFP SDK released after that date, you should provide an explicit base_image argument and ensure your component works as intended on Python 3.10."
            ),
            FutureWarning,
            stacklevel=2,
        )

    component_image = base_image

    if target_image:
        component_image = target_image
        command, args = _get_command_and_args_for_containerized_component(
            function_name=func.__name__,)
    else:
        command, args = _get_command_and_args_for_lightweight_component(
            func=func)

    component_spec = extract_component_interface(func)
    component_spec.implementation = structures.Implementation(
        container=structures.ContainerSpecImplementation(
            image=component_image,
            command=packages_to_install_command + command,
            args=args,
        ))

    module_path = pathlib.Path(inspect.getsourcefile(func))
    module_path.resolve()

    component_name = _python_function_name_to_component_name(func.__name__)
    component_info = ComponentInfo(
        name=component_name,
        function_name=func.__name__,
        func=func,
        target_image=target_image,
        module_path=module_path,
        component_spec=component_spec,
        output_component_file=output_component_file,
        base_image=base_image,
        packages_to_install=packages_to_install,
        pip_index_urls=pip_index_urls,
        pip_trusted_hosts=pip_trusted_hosts)

    if REGISTERED_MODULES is not None:
        REGISTERED_MODULES[component_name] = component_info

    if output_component_file:
        component_spec.save_to_component_yaml(output_component_file)

    return python_component.PythonComponent(
        component_spec=component_spec, python_func=func)


def make_input_for_parameterized_container_component_function(
    name: str, annotation: Union[Type[List[artifact_types.Artifact]],
                                 Type[artifact_types.Artifact]]
) -> Union[placeholders.Placeholder, container_component_artifact_channel
           .ContainerComponentArtifactChannel]:
    if type_annotations.is_artifact_wrapped_in_Input(annotation):

        if type_annotations.is_list_of_artifacts(annotation.__origin__):
            return placeholders.InputListOfArtifactsPlaceholder(name)
        else:
            return container_component_artifact_channel.ContainerComponentArtifactChannel(
                io_type='input', var_name=name)

    elif type_annotations.is_artifact_wrapped_in_Output(annotation):

        if type_annotations.is_list_of_artifacts(annotation.__origin__):
            return placeholders.OutputListOfArtifactsPlaceholder(name)
        else:
            return container_component_artifact_channel.ContainerComponentArtifactChannel(
                io_type='output', var_name=name)

    elif isinstance(
            annotation,
        (type_annotations.OutputAnnotation, type_annotations.OutputPath)):
        return placeholders.OutputParameterPlaceholder(name)

    else:
        placeholder = placeholders.InputValuePlaceholder(name)
        # small hack to encode the runtime value's type for a custom json.dumps function
        if (annotation == task_final_status.PipelineTaskFinalStatus or
                type_utils.is_task_final_status_type(annotation)):
            placeholder._ir_type = 'STRUCT'
        else:
            placeholder._ir_type = type_utils.get_parameter_type_name(
                annotation)
        return placeholder


def create_container_component_from_func(
        func: Callable) -> container_component_class.ContainerComponent:
    """Implementation for the @container_component decorator.

    The decorator is defined under container_component_decorator.py. See
    the decorator for the canonical documentation for this function.
    """

    component_spec = extract_component_interface(func, containerized=True)
    signature = inspect.signature(func)
    parameters = list(signature.parameters.values())
    arg_list = []
    for parameter in parameters:
        parameter_type = type_annotations.maybe_strip_optional_from_annotation(
            parameter.annotation)
        arg_list.append(
            make_input_for_parameterized_container_component_function(
                parameter.name, parameter_type))

    container_spec = func(*arg_list)
    container_spec_implementation = structures.ContainerSpecImplementation.from_container_spec(
        container_spec)
    component_spec.implementation = structures.Implementation(
        container_spec_implementation)
    component_spec._validate_placeholders()
    return container_component_class.ContainerComponent(component_spec, func)


def create_graph_component_from_func(
    func: Callable,
    name: Optional[str] = None,
    description: Optional[str] = None,
    display_name: Optional[str] = None,
    pipeline_config: pipeline_config.PipelineConfig = None,
) -> graph_component.GraphComponent:
    """Implementation for the @pipeline decorator.

    The decorator is defined under pipeline_context.py. See the
    decorator for the canonical documentation for this function.
    """

    component_spec = extract_component_interface(
        func,
        description=description,
        name=name,
    )
    return graph_component.GraphComponent(
        component_spec=component_spec,
        pipeline_func=func,
        display_name=display_name,
        pipeline_config=pipeline_config,
    )


def get_pipeline_description(
    decorator_description: Union[str, None],
    docstring: docstring_parser.Docstring,
) -> Optional[str]:
    """Obtains the correct pipeline description from the pipeline decorator's
    description argument and the parsed docstring.

    Gives precedence to the decorator argument.
    """
    if decorator_description:
        return decorator_description

    short_description = docstring.short_description
    long_description = docstring.long_description
    docstring_description = short_description + '\n' + long_description if (
        short_description and long_description) else short_description
    return docstring_description.strip() if docstring_description else None



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for kfp.dsl.pipeline_task."""

import textwrap
import unittest

from absl.testing import parameterized
from kfp import dsl
from kfp.dsl import pipeline_task
from kfp.dsl import placeholders
from kfp.dsl import structures

V2_YAML = textwrap.dedent("""\
components:
  comp-component1:
    executorLabel: exec-component1
    inputDefinitions:
      parameters:
        input1:
          parameterType: STRING
    outputDefinitions:
      artifacts:
        output1:
          artifactType:
            schemaTitle: system.Artifact
            schemaVersion: 0.0.1
deploymentSpec:
  executors:
    exec-component1:
      container:
        args:
        - '{{$.inputs.parameters[''input1'']}}'
        - '{{$.outputs.artifacts[''output1''].path}}'
        command:
        - sh
        - -c
        - echo "$0" >> "$1"
        image: alpine
pipelineInfo:
  name: component1
root:
  dag:
    tasks:
      component1:
        cachingOptions:
          enableCache: true
        componentRef:
          name: comp-component1
        inputs:
          parameters:
            input1:
              componentInputParameter: input1
        taskInfo:
          name: component1
  inputDefinitions:
    parameters:
      input1:
        parameterType: STRING
schemaVersion: 2.1.0
sdkVersion: kfp-2.0.0-alpha.2
""")


class PipelineTaskTest(parameterized.TestCase):

    def test_create_pipeline_task_valid(self):
        expected_component_spec = structures.ComponentSpec(
            name='component1',
            implementation=structures.Implementation(
                container=structures.ContainerSpecImplementation(
                    image='alpine',
                    command=['sh', '-c', 'echo "$0" >> "$1"'],
                    args=[
                        placeholders.InputValuePlaceholder(
                            input_name='input1')._to_string(),
                        placeholders.OutputPathPlaceholder(
                            output_name='output1')._to_string(),
                    ],
                )),
            inputs={
                'input1': structures.InputSpec(type='String'),
            },
            outputs={
                'output1': structures.OutputSpec(type='system.Artifact@0.0.1'),
            },
        )
        expected_task_spec = structures.TaskSpec(
            name='component1',
            inputs={'input1': 'value'},
            dependent_tasks=[],
            component_ref='component1',
        )
        expected_container_spec = structures.ContainerSpecImplementation(
            image='alpine',
            command=['sh', '-c', 'echo "$0" >> "$1"'],
            args=[
                "{{$.inputs.parameters['input1']}}",
                "{{$.outputs.artifacts['output1'].path}}",
            ],
        )

        task = pipeline_task.PipelineTask(
            component_spec=structures.ComponentSpec.from_yaml_documents(
                V2_YAML),
            args={'input1': 'value'},
        )
        self.assertEqual(task._task_spec, expected_task_spec)
        self.assertEqual(task.component_spec, expected_component_spec)
        self.assertEqual(task.container_spec, expected_container_spec)

    def test_create_pipeline_task_invalid_wrong_input(self):
        with self.assertRaisesRegex(
                ValueError,
                "Component 'component1' got an unexpected input: 'input0'."):
            task = pipeline_task.PipelineTask(
                component_spec=structures.ComponentSpec.from_yaml_documents(
                    V2_YAML),
                args={
                    'input1': 'value',
                    'input0': 'abc',
                },
            )

    def test_set_caching_options(self):
        task = pipeline_task.PipelineTask(
            component_spec=structures.ComponentSpec.from_yaml_documents(
                V2_YAML),
            args={'input1': 'value'},
        )
        task.set_caching_options(False)
        self.assertEqual(False, task._task_spec.enable_caching)

    @parameterized.parameters(
        {
            'cpu': '123',
            'expected_cpu': '123',
        },
        {
            'cpu': '123m',
            'expected_cpu': '123m',
        },
        {
            'cpu': '123.0',
            'expected_cpu': '123.0',
        },
        {
            'cpu': '123.0m',
            'expected_cpu': '123.0m',
        },
    )
    def test_set_valid_cpu_request_limit(self, cpu: str, expected_cpu: str):
        task = pipeline_task.PipelineTask(
            component_spec=structures.ComponentSpec.from_yaml_documents(
                V2_YAML),
            args={'input1': 'value'},
        )
        task.set_cpu_request(cpu)
        self.assertEqual(expected_cpu,
                         task.container_spec.resources.cpu_request)
        task.set_cpu_limit(cpu)
        self.assertEqual(expected_cpu, task.container_spec.resources.cpu_limit)

    @parameterized.parameters(
        {
            'gpu_limit': '1',
            'expected_gpu_number': '1',
        },)
    def test_set_valid_gpu_limit(self, gpu_limit: str,
                                 expected_gpu_number: str):
        task = pipeline_task.PipelineTask(
            component_spec=structures.ComponentSpec.from_yaml_documents(
                V2_YAML),
            args={'input1': 'value'},
        )
        with self.assertWarnsRegex(
                DeprecationWarning,
                "'set_gpu_limit' is deprecated. Please use 'set_accelerator_limit' instead."
        ):
            task.set_gpu_limit(gpu_limit)
        self.assertEqual(expected_gpu_number,
                         task.container_spec.resources.accelerator_count)

    def test_add_valid_node_selector_constraint(self):
        task = pipeline_task.PipelineTask(
            component_spec=structures.ComponentSpec.from_yaml_documents(
                V2_YAML),
            args={'input1': 'value'},
        )
        with self.assertWarnsRegex(
                DeprecationWarning,
                "'add_node_selector_constraint' is deprecated. Please use 'set_accelerator_type' instead."
        ):
            task.add_node_selector_constraint('TPU_V3')
        self.assertEqual(task.container_spec.resources.accelerator_type,
                         'TPU_V3')

    @parameterized.parameters(
        {
            'limit': '1',
            'expected_limit': '1',
        },
        {
            'limit': 1,
            'expected_limit': '1',
        },
        {
            'limit': 16,
            'expected_limit': '16',
        },
    )
    def test_set_accelerator_limit(self, limit, expected_limit):
        task = pipeline_task.PipelineTask(
            component_spec=structures.ComponentSpec.from_yaml_documents(
                V2_YAML),
            args={'input1': 'value'},
        )

        task.set_accelerator_limit(limit)
        self.assertEqual(expected_limit,
                         task.container_spec.resources.accelerator_count)

    @parameterized.parameters(
        {
            'memory': '1E',
            'expected_memory': '1E',
        },
        {
            'memory': '15Ei',
            'expected_memory': '15Ei',
        },
        {
            'memory': '2P',
            'expected_memory': '2P',
        },
        {
            'memory': '25Pi',
            'expected_memory': '25Pi',
        },
        {
            'memory': '3T',
            'expected_memory': '3T',
        },
        {
            'memory': '35Ti',
            'expected_memory': '35Ti',
        },
        {
            'memory': '4G',
            'expected_memory': '4G',
        },
        {
            'memory': '45Gi',
            'expected_memory': '45Gi',
        },
        {
            'memory': '5M',
            'expected_memory': '5M',
        },
        {
            'memory': '55Mi',
            'expected_memory': '55Mi',
        },
        {
            'memory': '6K',
            'expected_memory': '6K',
        },
        {
            'memory': '65Ki',
            'expected_memory': '65Ki',
        },
        {
            'memory': '7000',
            'expected_memory': '7000',
        },
    )
    def test_set_memory_limit(self, memory: str, expected_memory: str):
        task = pipeline_task.PipelineTask(
            component_spec=structures.ComponentSpec.from_yaml_documents(
                V2_YAML),
            args={'input1': 'value'},
        )
        task.set_memory_request(memory)
        self.assertEqual(expected_memory,
                         task.container_spec.resources.memory_request)
        task.set_memory_limit(memory)
        self.assertEqual(expected_memory,
                         task.container_spec.resources.memory_limit)

    def test_set_accelerator_type_with_type_only(self):
        task = pipeline_task.PipelineTask(
            component_spec=structures.ComponentSpec.from_yaml_documents(
                V2_YAML),
            args={'input1': 'value'},
        )
        task.set_accelerator_type('NVIDIA_TESLA_K80')
        self.assertEqual(
            structures.ResourceSpec(
                accelerator_type='NVIDIA_TESLA_K80', accelerator_count='1'),
            task.container_spec.resources)

    def test_set_accelerator_type_with_accelerator_count(self):
        task = pipeline_task.PipelineTask(
            component_spec=structures.ComponentSpec.from_yaml_documents(
                V2_YAML),
            args={'input1': 'value'},
        )
        task.set_accelerator_limit('4').set_accelerator_type('TPU_V3')
        self.assertEqual(
            structures.ResourceSpec(
                accelerator_type='TPU_V3', accelerator_count='4'),
            task.container_spec.resources)

    def test_set_env_variable(self):
        task = pipeline_task.PipelineTask(
            component_spec=structures.ComponentSpec.from_yaml_documents(
                V2_YAML),
            args={'input1': 'value'},
        )
        task.set_env_variable('env_name', 'env_value')
        self.assertEqual({'env_name': 'env_value'}, task.container_spec.env)

    def test_set_display_name(self):
        task = pipeline_task.PipelineTask(
            component_spec=structures.ComponentSpec.from_yaml_documents(
                V2_YAML),
            args={'input1': 'value'},
        )
        task.set_display_name('test_name')
        self.assertEqual('test_name', task._task_spec.display_name)

    def test_set_cpu_limit_on_pipeline_should_raise(self):

        @dsl.component
        def comp():
            print('hello')

        @dsl.pipeline
        def graph():
            comp()
            comp()

        with self.assertRaisesRegex(
                ValueError,
                r'set_cpu_limit can only be used on single-step components'):

            @dsl.pipeline
            def my_pipeline():
                graph().set_cpu_limit('1')


class TestPlatformSpecificFunctionality(unittest.TestCase):

    def test_platform_config_to_platform_spec(self):

        @dsl.component
        def comp():
            pass

        @dsl.pipeline
        def my_pipeline():
            t = comp()
            t.platform_config = {'platform1': {'feature': [1, 2, 3]}}
            with self.assertRaisesRegex(
                    ValueError,
                    r"Can only access '\.platform_spec' property on a tasks created from pipelines\. Use '\.platform_config' for tasks created from primitive components\."
            ):
                t.platform_spec


class TestTaskInFinalState(unittest.TestCase):
    """Tests PipelineTask in the state FINAL.

    Many properties and methods will be blocked.

    Also tests that the .output and .outputs behavior behaves as expected when the outputs are values, not placeholders, as will be the case when PipelineTask is in the state FINAL.
    """

    def test_output_property(self):
        task = pipeline_task.PipelineTask(
            component_spec=structures.ComponentSpec.from_yaml_documents(
                V2_YAML),
            args={'input1': 'value'},
        )
        task.state = pipeline_task.TaskState.FINAL
        task._outputs = {'Output': 1}
        self.assertEqual(task.output, 1)
        self.assertEqual(task.outputs['Output'], 1)

    def test_outputs_property(self):
        task = pipeline_task.PipelineTask(
            component_spec=structures.ComponentSpec.from_yaml_documents(
                V2_YAML),
            args={'input1': 'value'},
        )
        task.state = pipeline_task.TaskState.FINAL
        task._outputs = {
            'int_output':
                1,
            'str_output':
                'foo',
            'dataset_output':
                dsl.Dataset(
                    name='dataset_output',
                    uri='foo/bar/dataset_output',
                    metadata={'key': 'value'})
        }
        self.assertEqual(task.outputs['int_output'], 1)
        self.assertEqual(task.outputs['str_output'], 'foo')
        assert_artifacts_equal(
            self,
            task.outputs['dataset_output'],
            dsl.Dataset(
                name='dataset_output',
                uri='foo/bar/dataset_output',
                metadata={'key': 'value'}),
        )

    def test_platform_spec_property(self):
        task = pipeline_task.PipelineTask(
            component_spec=structures.ComponentSpec.from_yaml_documents(
                V2_YAML),
            args={'input1': 'value'},
        )
        task.state = pipeline_task.TaskState.FINAL
        with self.assertRaisesRegex(
                Exception,
                r'Platform-specific features are not supported for local execution\.'
        ):
            task.platform_spec

    def test_name_property(self):
        task = pipeline_task.PipelineTask(
            component_spec=structures.ComponentSpec.from_yaml_documents(
                V2_YAML),
            args={'input1': 'value'},
        )
        task.state = pipeline_task.TaskState.FINAL
        self.assertEqual(task.name, 'component1')

    def test_inputs_property(self):
        task = pipeline_task.PipelineTask(
            component_spec=structures.ComponentSpec.from_yaml_documents(
                V2_YAML),
            args={'input1': 'value'},
        )
        task.state = pipeline_task.TaskState.FINAL
        self.assertEqual(task.inputs, {'input1': 'value'})

    def test_dependent_tasks_property(self):
        task = pipeline_task.PipelineTask(
            component_spec=structures.ComponentSpec.from_yaml_documents(
                V2_YAML),
            args={'input1': 'value'},
        )
        task.state = pipeline_task.TaskState.FINAL
        with self.assertRaisesRegex(
                Exception,
                r'Task has no dependent tasks since it is executed independently\.'
        ):
            task.dependent_tasks

    def test_sampling_of_task_configuration_methods(self):
        task = pipeline_task.PipelineTask(
            component_spec=structures.ComponentSpec.from_yaml_documents(
                V2_YAML),
            args={'input1': 'value'},
        )
        task.state = pipeline_task.TaskState.FINAL
        with self.assertRaisesRegex(
                Exception,
                r"Task configuration methods are not supported for local execution\. Got call to '\.set_caching_options\(\)'\."
        ):
            task.set_caching_options(enable_caching=True)
        with self.assertRaisesRegex(
                Exception,
                r"Task configuration methods are not supported for local execution\. Got call to '\.set_env_variable\(\)'\."
        ):
            task.set_env_variable(name='foo', value='BAR')
        with self.assertRaisesRegex(
                Exception,
                r"Task configuration methods are not supported for local execution\. Got call to '\.ignore_upstream_failure\(\)'\."
        ):
            task.ignore_upstream_failure()


def assert_artifacts_equal(
    test_class: unittest.TestCase,
    a1: dsl.Artifact,
    a2: dsl.Artifact,
) -> None:
    test_class.assertEqual(a1.name, a2.name)
    test_class.assertEqual(a1.uri, a2.uri)
    test_class.assertEqual(a1.metadata, a2.metadata)
    test_class.assertEqual(a1.schema_title, a2.schema_title)
    test_class.assertEqual(a1.schema_version, a2.schema_version)
    test_class.assertIsInstance(a1, type(a2))


if __name__ == '__main__':
    unittest.main()



# Copyright 2021-2022 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functions for loading components from compiled YAML."""

from typing import Optional, Tuple

from kfp.dsl import structures
from kfp.dsl import yaml_component
import requests


def load_component_from_text(text: str) -> yaml_component.YamlComponent:
    """Loads a component from text.

    Args:
        text (str): Component YAML text.

    Returns:
        Component loaded from YAML.
    """
    return yaml_component.YamlComponent(
        component_spec=structures.ComponentSpec.from_yaml_documents(text),
        component_yaml=text)


def load_component_from_file(file_path: str) -> yaml_component.YamlComponent:
    """Loads a component from a file.

    Args:
        file_path (str): Filepath to a YAML component.

    Returns:
        Component loaded from YAML.

    Example:
      ::

        from kfp import components

        components.load_component_from_file('~/path/to/pipeline.yaml')
    """
    with open(file_path, 'r') as component_stream:
        return load_component_from_text(component_stream.read())


def load_component_from_url(
        url: str,
        auth: Optional[Tuple[str, str]] = None) -> yaml_component.YamlComponent:
    """Loads a component from a URL.

    Args:
        url (str): URL to a YAML component.
        auth (Tuple[str, str], optional): A ``('<username>', '<password>')`` tuple of authentication credentials necessary for URL access. See `Requests Authorization <https://requests.readthedocs.io/en/latest/user/authentication/#authentication>`_ for more information.

    Returns:
        Component loaded from YAML.

    Example:
      ::

        from kfp import components

        components.load_component_from_url('https://raw.githubusercontent.com/kubeflow/pipelines/7b49eadf621a9054e1f1315c86f95fb8cf8c17c3/sdk/python/kfp/compiler/test_data/components/identity.yaml')

        components.load_component_from_url('gs://path/to/pipeline.yaml')
    """
    if url is None:
        raise ValueError('url must be a string.')

    if url.startswith('gs://'):
        #Replacing the gs:// URI with https:// URI (works for public objects)
        url = 'https://storage.googleapis.com/' + url[len('gs://'):]

    resp = requests.get(url, auth=auth)
    resp.raise_for_status()

    return load_component_from_text(resp.content.decode('utf-8'))



# Copyright 2023 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for task_handler_interface_test.py."""
import unittest

from kfp.local import task_handler_interface


class TestTaskHandler(unittest.TestCase):

    def test_incomplete_concrete_class(self):

        class IncompleteTaskHandler(
                task_handler_interface.ITaskHandler,):

            pass

        with self.assertRaisesRegex(
                TypeError,
                r"Can't instantiate abstract class IncompleteTaskHandler"):
            IncompleteTaskHandler()


if __name__ == '__main__':
    unittest.main()



# Copyright 2023 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from unittest.mock import MagicMock
from unittest.mock import patch

from absl.testing import parameterized
from kfp.client import auth


class TestAuth(parameterized.TestCase):

    def test_is_ipython_return_false(self):
        mock = MagicMock()
        with patch.dict('sys.modules', IPython=mock):
            mock.get_ipython.return_value = None
            self.assertFalse(auth.is_ipython())

    def test_is_ipython_return_true(self):
        mock = MagicMock()
        with patch.dict('sys.modules', IPython=mock):
            mock.get_ipython.return_value = 'Something'
            self.assertTrue(auth.is_ipython())

    def test_is_ipython_should_raise_error(self):
        mock = MagicMock()
        with patch.dict('sys.modules', mock):
            mock.side_effect = ImportError
            self.assertFalse(auth.is_ipython())

    @patch('builtins.input', lambda *args:
           'https://oauth2.example.com/auth?code=4/P7q7W91a-oMsCeLvIaQm6bTrgtp7'
          )
    @patch('kfp.client.auth.is_ipython', lambda *args: True)
    @patch.dict(os.environ, dict(), clear=True)
    def test_get_auth_code_from_ipython(self):
        token, redirect_uri = auth.get_auth_code('sample-client-id')
        self.assertEqual(token, '4/P7q7W91a-oMsCeLvIaQm6bTrgtp7')
        self.assertEqual(redirect_uri, 'http://localhost:9901')

    @patch('builtins.input', lambda *args:
           'https://oauth2.example.com/auth?code=4/P7q7W91a-oMsCeLvIaQm6bTrgtp7'
          )
    @patch('kfp.client.auth.is_ipython', lambda *args: False)
    @patch.dict(os.environ, {'SSH_CONNECTION': 'ENABLED'}, clear=True)
    def test_get_auth_code_from_remote_connection(self):
        token, redirect_uri = auth.get_auth_code('sample-client-id')
        self.assertEqual(token, '4/P7q7W91a-oMsCeLvIaQm6bTrgtp7')
        self.assertEqual(redirect_uri, 'http://localhost:9901')

    @patch('builtins.input', lambda *args:
           'https://oauth2.example.com/auth?code=4/P7q7W91a-oMsCeLvIaQm6bTrgtp7'
          )
    @patch('kfp.client.auth.is_ipython', lambda *args: False)
    @patch.dict(os.environ, {'SSH_CLIENT': 'ENABLED'}, clear=True)
    def test_get_auth_code_from_remote_client(self):
        token, redirect_uri = auth.get_auth_code('sample-client-id')
        self.assertEqual(token, '4/P7q7W91a-oMsCeLvIaQm6bTrgtp7')
        self.assertEqual(redirect_uri, 'http://localhost:9901')

    @patch('builtins.input', lambda *args: 'https://oauth2.example.com/auth')
    @patch('kfp.client.auth.is_ipython', lambda *args: False)
    @patch.dict(os.environ, {'SSH_CLIENT': 'ENABLED'}, clear=True)
    def test_get_auth_code_from_remote_client_missing_code(self):
        self.assertRaises(KeyError, auth.get_auth_code, 'sample-client-id')

    @patch('kfp.client.auth.get_auth_response_local', lambda *args:
           'https://oauth2.example.com/auth?code=4/P7q7W91a-oMsCeLvIaQm6bTrgtp7'
          )
    @patch('kfp.client.auth.is_ipython', lambda *args: False)
    @patch.dict(os.environ, dict(), clear=True)
    def test_get_auth_code_from_local(self):
        token, redirect_uri = auth.get_auth_code('sample-client-id')
        self.assertEqual(token, '4/P7q7W91a-oMsCeLvIaQm6bTrgtp7')
        self.assertEqual(redirect_uri, 'http://localhost:9901')

    @patch('kfp.client.auth.get_auth_response_local', lambda *args: None)
    @patch('kfp.client.auth.is_ipython', lambda *args: False)
    @patch.dict(os.environ, dict(), clear=True)
    def test_get_auth_code_from_local_empty_response(self):
        self.assertRaises(ValueError, auth.get_auth_code, 'sample-client-id')

    @patch('kfp.client.auth.get_auth_response_local',
           lambda *args: 'this-is-an-invalid-response')
    @patch('kfp.client.auth.is_ipython', lambda *args: False)
    @patch.dict(os.environ, dict(), clear=True)
    def test_get_auth_code_from_local_invalid_response(self):
        self.assertRaises(KeyError, auth.get_auth_code, 'sample-client-id')



# Copyright 2024 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for io.py."""

import unittest

from kfp import dsl
from kfp.local import io


class IOStoreTest(unittest.TestCase):

    def test_task_not_found(self):
        store = io.IOStore()
        with self.assertRaisesRegex(
                ValueError,
                r"Tried to get output 'foo' from task 'my-task', but task 'my-task' not found\."
        ):
            store.get_task_output('my-task', 'foo')

    def test_output_not_found(self):
        store = io.IOStore()
        store.put_task_output('my-task', 'bar', 'baz')
        with self.assertRaisesRegex(
                ValueError,
                r"Tried to get output 'foo' from task 'my-task', but task 'my-task' has no output named 'foo'\."
        ):
            store.get_task_output('my-task', 'foo')

    def test_parent_input_not_found(self):
        store = io.IOStore()
        with self.assertRaisesRegex(
                ValueError, r"Parent pipeline input argument 'foo' not found."):
            store.get_parent_input('foo')

    def test_put_and_get_task_output(self):
        store = io.IOStore()
        store.put_task_output('my-task', 'foo', 'bar')
        store.put_task_output('my-task', 'baz', 'bat')
        self.assertEqual(
            store.get_task_output('my-task', 'foo'),
            'bar',
        )
        self.assertEqual(
            store.get_task_output('my-task', 'baz'),
            'bat',
        )
        # test getting doesn't remove by getting twice
        self.assertEqual(
            store.get_task_output('my-task', 'baz'),
            'bat',
        )

    def test_put_and_get_parent_input(self):
        store = io.IOStore()
        store.put_parent_input('foo', 'bar')
        store.put_parent_input('baz', 'bat')
        self.assertEqual(
            store.get_parent_input('foo'),
            'bar',
        )
        self.assertEqual(
            store.get_parent_input('baz'),
            'bat',
        )
        # test getting doesn't remove by getting twice
        self.assertEqual(
            store.get_parent_input('baz'),
            'bat',
        )

    def test_put_and_get_task_output_with_artifact(self):
        artifact = dsl.Artifact(
            name='foo', uri='/my/uri', metadata={'foo': 'bar'})
        store = io.IOStore()
        store.put_task_output('my-task', 'foo', artifact)
        self.assertEqual(
            store.get_task_output('my-task', 'foo'),
            artifact,
        )


if __name__ == '__main__':
    unittest.main()



# Copyright 2024 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Object for storing task outputs in-memory during local execution."""

import collections
from typing import Any, Dict


class IOStore:

    def __init__(self):

        self._task_output_data: Dict[str,
                                     Dict[str,
                                          Any]] = collections.defaultdict(dict)
        self._parent_input_data: Dict[str, Any] = {}

    def put_parent_input(
        self,
        key: str,
        value: Any,
    ) -> None:
        """Persist the value of a parent component (i.e., parent pipeline)
        input.

        Args:
            key: Parent component input name.
            value: Value associated with key.
        """
        self._parent_input_data[key] = value

    def get_parent_input(
        self,
        key: str,
    ) -> None:
        """Get the value of the parent component (i.e., parent pipeline) input
        named key.

        Args:
            key: Parent component input name.

        Returns:
            The output value.
        """
        if key in self._parent_input_data:
            return self._parent_input_data[key]
        raise ValueError(f"Parent pipeline input argument '{key}' not found.")

    def put_task_output(
        self,
        task_name: str,
        key: str,
        value: Any,
    ) -> None:
        """Persist the value of an upstream task output.

        Args:
            task_name: Upstream task name.
            key: Output name.
            value: Value associated with key.
        """
        self._task_output_data[task_name][key] = value

    def get_task_output(
        self,
        task_name: str,
        key: str,
    ) -> Any:
        """Get the value of an upstream task output.

        Args:
            task_name: Upstream task name.
            key: Output name.

        Returns:
            The output value.
        """
        common_exception_string = f"Tried to get output '{key}' from task '{task_name}'"
        if task_name in self._task_output_data:
            outputs = self._task_output_data[task_name]
        else:
            raise ValueError(
                f"{common_exception_string}, but task '{task_name}' not found.")

        if key in outputs:
            return outputs[key]
        else:
            raise ValueError(
                f"{common_exception_string}, but task '{task_name}' has no output named '{key}'."
            )



# Copyright 2023 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
import sys
import unittest
from unittest import mock


@mock.patch.object(sys, 'version_info', new=(3, 7, 12, 'final', 0))
class TestPythonEOLWarning(unittest.TestCase):

    def test(self):
        mod = importlib.import_module('kfp')

        with self.assertWarnsRegex(
                FutureWarning,
                r'KFP will drop support for Python 3.9 on Oct 1, 2025. To use new versions of the KFP SDK after that date, you will need to upgrade to Python >= 3.10. See https://devguide.python.org/versions/ for more details.'
        ):
            # simulate first import from kfp
            importlib.reload(mod)


if __name__ == '__main__':
    unittest.main()



# Copyright 2024 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for pipeline_orchestrator.py."""
import functools
import io as stdlib_io
import os
from typing import NamedTuple
import unittest
from unittest import mock

from kfp import dsl
from kfp import local
from kfp.dsl import Dataset
from kfp.dsl import Input
from kfp.dsl import Model
from kfp.dsl import Output
from kfp.dsl import pipeline_task
from kfp.local import testing_utilities
import pytest

ROOT_FOR_TESTING = './testing_root'


@pytest.fixture(autouse=True)
def set_packages_for_test_classes(monkeypatch, request):
    if request.cls.__name__ in {
            'TestRunLocalPipeline',
            'TestFstringContainerComponent',
    }:
        root_dir = os.path.dirname(
            os.path.dirname(
                os.path.dirname(os.path.dirname(os.path.dirname(__file__)))))
        kfp_pipeline_spec_path = os.path.join(root_dir, 'api', 'v2alpha1',
                                              'python')
        original_dsl_component = dsl.component
        monkeypatch.setattr(
            dsl, 'component',
            functools.partial(
                original_dsl_component,
                packages_to_install=[kfp_pipeline_spec_path]))


class TestRunLocalPipeline(testing_utilities.LocalRunnerEnvironmentTestCase):

    def assert_output_dir_contents(
        self,
        expected_dirs_in_pipeline_root: int,
        expected_files_in_pipeline_dir: int,
    ) -> None:
        # check that output files are correctly nested
        # and only one directory for the outer pipeline in pipeline root
        actual_dirs_in_pipeline_root = os.listdir(ROOT_FOR_TESTING)
        self.assertLen(
            actual_dirs_in_pipeline_root,
            expected_dirs_in_pipeline_root,
        )

        # and check that each task has a directory
        actual_contents_of_pipeline_dir = os.listdir(
            os.path.join(
                ROOT_FOR_TESTING,
                actual_dirs_in_pipeline_root[0],
            ))
        self.assertLen(
            actual_contents_of_pipeline_dir,
            expected_files_in_pipeline_dir,
        )

    def test_must_initialize(self):

        @dsl.component
        def identity(string: str) -> str:
            return string

        @dsl.pipeline
        def my_pipeline():
            identity(string='foo')

        with self.assertRaisesRegex(
                RuntimeError,
                r"Local environment not initialized\. Please run 'kfp\.local\.init\(\)' before executing tasks locally\."
        ):
            my_pipeline()

    def test_no_io(self):
        local.init(local.SubprocessRunner(), pipeline_root=ROOT_FOR_TESTING)

        @dsl.component()
        def pass_op():
            pass

        @dsl.pipeline
        def my_pipeline():
            pass_op()
            pass_op()

        result = my_pipeline()
        self.assertIsInstance(result, pipeline_task.PipelineTask)
        self.assertEqual(result.outputs, {})
        self.assert_output_dir_contents(1, 2)

    def test_missing_args(self):
        local.init(local.SubprocessRunner())

        @dsl.component
        def identity(string: str) -> str:
            return string

        @dsl.pipeline
        def my_pipeline(string: str) -> str:
            t1 = identity(string=string)
            t2 = identity(string=t1.output)
            return t2.output

        with self.assertRaisesRegex(
                TypeError,
                r'my-pipeline\(\) missing 1 required argument: string\.'):
            my_pipeline()

    def test_single_return(self):
        local.init(local.SubprocessRunner(), pipeline_root=ROOT_FOR_TESTING)

        @dsl.component
        def identity(string: str) -> str:
            return string

        @dsl.pipeline
        def my_pipeline(string: str = 'text') -> str:
            t1 = identity(string=string)
            t2 = identity(string=t1.output)
            return t2.output

        task = my_pipeline()
        self.assertEqual(task.output, 'text')
        self.assert_output_dir_contents(1, 2)

    def test_can_run_loaded_pipeline(self):
        local.init(local.SubprocessRunner(), pipeline_root=ROOT_FOR_TESTING)

        @dsl.component
        def identity(string: str) -> str:
            return string

        @dsl.pipeline
        def my_pipeline(string: str = 'text') -> str:
            t1 = identity(string=string)
            t2 = identity(string=t1.output)
            return t2.output

        my_pipeline_loaded = testing_utilities.compile_and_load_component(
            my_pipeline)

        task = my_pipeline_loaded(string='foo')
        self.assertEqual(task.output, 'foo')
        self.assert_output_dir_contents(1, 2)

    def test_all_param_io(self):
        local.init(local.SubprocessRunner(), pipeline_root=ROOT_FOR_TESTING)

        # tests all I/O types with:
        # - use of component args/defaults
        # - use of pipeline args/defaults
        # - passing pipeline args to first run component and not first run component
        # - passing args from pipeline param, upstream output, and constant
        # - pipeline surfacing outputs from last run component and not last run component

        @dsl.component
        def many_parameter_component(
            a_float: float,
            a_boolean: bool,
            a_dict: dict,
            a_string: str = 'default',
            an_integer: int = 12,
            a_list: list = ['item1', 'item2'],
        ) -> NamedTuple(
                'outputs',
                a_string=str,
                a_float=float,
                an_integer=int,
                a_boolean=bool,
                a_list=list,
                a_dict=dict,
        ):
            outputs = NamedTuple(
                'outputs',
                a_string=str,
                a_float=float,
                an_integer=int,
                a_boolean=bool,
                a_list=list,
                a_dict=dict,
            )
            return outputs(
                a_string=a_string,
                a_float=a_float,
                an_integer=an_integer,
                a_boolean=a_boolean,
                a_list=a_list,
                a_dict=a_dict,
            )

        @dsl.pipeline
        def my_pipeline(
            flt: float,
            boolean: bool,
            dictionary: dict = {'foo': 'bar'},
        ) -> NamedTuple(
                'outputs',
                another_string=str,
                another_float=float,
                another_integer=int,
                another_boolean=bool,
                another_list=list,
                another_dict=dict,
        ):

            t1 = many_parameter_component(
                a_float=flt,
                a_boolean=True,
                a_dict={'baz': 'bat'},
                an_integer=10,
            )
            t2 = many_parameter_component(
                a_float=t1.outputs['a_float'],
                a_dict=dictionary,
                a_boolean=boolean,
            )

            outputs = NamedTuple(
                'outputs',
                another_string=str,
                another_float=float,
                another_integer=int,
                another_boolean=bool,
                another_list=list,
                another_dict=dict,
            )
            return outputs(
                another_string=t1.outputs['a_string'],
                another_float=t1.outputs['a_float'],
                another_integer=t1.outputs['an_integer'],
                another_boolean=t2.outputs['a_boolean'],
                another_list=t2.outputs['a_list'],
                another_dict=t1.outputs['a_dict'],
            )

        task = my_pipeline(
            flt=2.718,
            boolean=False,
        )
        self.assertEqual(task.outputs['another_string'], 'default')
        self.assertEqual(task.outputs['another_float'], 2.718)
        self.assertEqual(task.outputs['another_integer'], 10)
        self.assertEqual(task.outputs['another_boolean'], False)
        self.assertEqual(task.outputs['another_list'], ['item1', 'item2'])
        self.assertEqual(task.outputs['another_dict'], {'baz': 'bat'})
        self.assert_output_dir_contents(1, 2)

    def test_artifact_io(self):
        local.init(local.SubprocessRunner(), pipeline_root=ROOT_FOR_TESTING)

        @dsl.component
        def make_dataset(content: str) -> Dataset:
            d = Dataset(uri=dsl.get_uri(), metadata={'framework': 'pandas'})
            with open(d.path, 'w') as f:
                f.write(content)
            return d

        @dsl.component
        def make_model(dataset: Input[Dataset], model: Output[Model]):
            with open(dataset.path) as f:
                content = f.read()
            with open(model.path, 'w') as f:
                f.write(content * 2)
            model.metadata['framework'] = 'tensorflow'
            model.metadata['dataset'] = dataset.metadata

        @dsl.pipeline
        def my_pipeline(content: str = 'string') -> Model:
            t1 = make_dataset(content=content)
            t2 = make_model(dataset=t1.output)
            return t2.outputs['model']

        task = my_pipeline(content='text')
        output_model = task.output
        self.assertIsInstance(output_model, Model)
        self.assertEqual(output_model.name, 'model')
        self.assertTrue(output_model.uri.endswith('/make-model/model'))
        self.assertEqual(output_model.metadata, {
            'framework': 'tensorflow',
            'dataset': {
                'framework': 'pandas'
            }
        })
        self.assert_output_dir_contents(1, 2)

    def test_input_artifact_constant_not_permitted(self):
        local.init(local.SubprocessRunner(), pipeline_root=ROOT_FOR_TESTING)

        @dsl.component
        def print_model(model: Input[Model]):
            print(model.name)
            print(model.uri)
            print(model.metadata)

        with self.assertRaisesRegex(
                ValueError,
                r"Input artifacts are not supported\. Got input artifact of type 'Model'\."
        ):

            @dsl.pipeline
            def my_pipeline():
                print_model(model=dsl.Model(name='model', uri='/foo/bar/model'))

    def test_importer(self):
        local.init(local.SubprocessRunner(), pipeline_root=ROOT_FOR_TESTING)

        @dsl.component
        def artifact_printer(a: Dataset):
            print(a)

        @dsl.component
        def identity(string: str) -> str:
            return string

        @dsl.pipeline
        def my_pipeline(greeting: str) -> Dataset:
            world_op = identity(string='world')
            message_op = identity(string='message')
            imp_op = dsl.importer(
                artifact_uri='/local/path/to/dataset',
                artifact_class=Dataset,
                metadata={
                    message_op.output: [greeting, world_op.output],
                })
            artifact_printer(a=imp_op.outputs['artifact'])
            return imp_op.outputs['artifact']

        task = my_pipeline(greeting='hello')
        output_model = task.output
        self.assertIsInstance(output_model, Dataset)
        self.assertEqual(output_model.name, 'artifact')
        self.assertEqual(output_model.uri, '/local/path/to/dataset')
        self.assertEqual(output_model.metadata, {
            'message': ['hello', 'world'],
        })
        # importer doesn't have an output directory
        self.assert_output_dir_contents(1, 3)

    def test_pipeline_in_pipeline_simple(self):
        local.init(local.SubprocessRunner(), pipeline_root=ROOT_FOR_TESTING)

        @dsl.component
        def identity(string: str) -> str:
            return string

        @dsl.pipeline
        def inner_pipeline() -> str:
            return identity(string='foo').output

        @dsl.pipeline
        def my_pipeline() -> str:
            return inner_pipeline().output

        task = my_pipeline()
        self.assertEqual(task.output, 'foo')
        self.assert_output_dir_contents(1, 1)

    def test_pipeline_in_pipeline_complex(self):
        local.init(local.SubprocessRunner(), pipeline_root=ROOT_FOR_TESTING)

        @dsl.component
        def square(x: float) -> float:
            return x**2

        @dsl.component
        def add(x: float, y: float) -> float:
            return x + y

        @dsl.component
        def square_root(x: float) -> float:
            return x**.5

        @dsl.component
        def convert_to_artifact(val: float) -> Dataset:
            dataset = Dataset(name='dataset', uri=dsl.get_uri())
            with open(dataset.path, 'w') as f:
                f.write(str(val))

        @dsl.component
        def convert_from_artifact(dataset: Dataset) -> float:
            with open(dataset.path) as f:
                return float(f.read())

        @dsl.pipeline
        def square_and_sum(a: float, b: float) -> Dataset:
            a_sq_task = square(x=a)
            b_sq_task = square(x=b)
            add_task = add(x=a_sq_task.output, y=b_sq_task.output)
            return convert_to_artifact(val=add_task.output).output

        @dsl.pipeline
        def pythagorean(a: float = 1.2, b: float = 1.2) -> float:
            sq_and_sum_task = square_and_sum(a=a, b=b)
            make_float_task = convert_from_artifact(
                dataset=sq_and_sum_task.output)
            return square_root(x=make_float_task.output).output

        @dsl.pipeline
        def pythagorean_then_add(
            side: float,
            addend: float = 42.24,
        ) -> float:
            t = pythagorean(a=side, b=1.2)
            return add(x=t.output, y=addend).output

        task = pythagorean_then_add(side=2.2)
        self.assertAlmostEqual(task.output, 44.745992817228334)
        self.assert_output_dir_contents(1, 7)

    def test_parallel_for_not_supported(self):
        local.init(local.SubprocessRunner())

        @dsl.component
        def pass_op():
            pass

        @dsl.pipeline
        def my_pipeline():
            with dsl.ParallelFor([1, 2, 3]):
                pass_op()

        with self.assertRaisesRegex(
                NotImplementedError,
                r"'dsl\.ParallelFor' is not supported by local pipeline execution\."
        ):
            my_pipeline()

    def test_condition_not_supported(self):
        local.init(local.SubprocessRunner())

        @dsl.component
        def pass_op():
            pass

        @dsl.pipeline
        def my_pipeline(x: str):
            with dsl.Condition(x == 'foo'):
                pass_op()

        with self.assertRaisesRegex(
                NotImplementedError,
                r"'dsl\.Condition' is not supported by local pipeline execution\."
        ):
            my_pipeline(x='bar')

    @mock.patch('sys.stdout', new_callable=stdlib_io.StringIO)
    def test_fails_with_raise_on_error_true(self, mock_stdout):
        local.init(local.SubprocessRunner(), raise_on_error=True)

        @dsl.component
        def raise_component():
            raise Exception('Error from raise_component.')

        @dsl.pipeline
        def my_pipeline():
            raise_component()

        with self.assertRaisesRegex(
                RuntimeError,
                r"Pipeline \x1b\[95m\'my-pipeline\'\x1b\[0m finished with status \x1b\[91mFAILURE\x1b\[0m\. Inner task failed: \x1b\[96m\'raise-component\'\x1b\[0m\.",
        ):
            my_pipeline()

        logged_output = mock_stdout.getvalue()
        # Logs should:
        # - log task failure trace
        # - log pipeline failure
        # - indicate which task the failure came from
        self.assertRegex(
            logged_output,
            r"raise Exception\('Error from raise_component\.'\)",
        )

    @mock.patch('sys.stdout', new_callable=stdlib_io.StringIO)
    def test_single_nested_fails_with_raise_on_error_true(self, mock_stdout):
        local.init(local.SubprocessRunner(), raise_on_error=True)

        @dsl.component
        def fail():
            raise Exception('Nested failure!')

        @dsl.pipeline
        def inner_pipeline():
            fail()

        @dsl.pipeline
        def my_pipeline():
            inner_pipeline()

        with self.assertRaisesRegex(
                RuntimeError,
                r"Pipeline \x1b\[95m\'my-pipeline\'\x1b\[0m finished with status \x1b\[91mFAILURE\x1b\[0m\. Inner task failed: \x1b\[96m\'fail\'\x1b\[0m inside \x1b\[96m\'inner-pipeline\'\x1b\[0m\.",
        ):
            my_pipeline()

        logged_output = mock_stdout.getvalue()
        self.assertRegex(
            logged_output,
            r"raise Exception\('Nested failure!'\)",
        )

    @mock.patch('sys.stdout', new_callable=stdlib_io.StringIO)
    def test_deeply_nested_fails_with_raise_on_error_true(self, mock_stdout):
        local.init(local.SubprocessRunner(), raise_on_error=True)

        @dsl.component
        def fail():
            raise Exception('Nested failure!')

        @dsl.pipeline
        def deep_pipeline():
            fail()

        @dsl.pipeline
        def mid_pipeline():
            deep_pipeline()

        @dsl.pipeline
        def outer_pipeline():
            mid_pipeline()

        with self.assertRaisesRegex(
                RuntimeError,
                r"Pipeline \x1b\[95m\'outer-pipeline\'\x1b\[0m finished with status \x1b\[91mFAILURE\x1b\[0m\. Inner task failed: \x1b\[96m\'fail\'\x1b\[0m\ inside \x1b\[96m\'deep-pipeline\'\x1b\[0m inside \x1b\[96m\'mid-pipeline\'\x1b\[0m\.",
        ):
            outer_pipeline()

        logged_output = mock_stdout.getvalue()
        self.assertRegex(
            logged_output,
            r"raise Exception\('Nested failure!'\)",
        )

    @mock.patch('sys.stdout', new_callable=stdlib_io.StringIO)
    def test_fails_with_raise_on_error_false(self, mock_stdout):
        local.init(local.SubprocessRunner(), raise_on_error=False)

        @dsl.component
        def raise_component():
            raise Exception('Error from raise_component.')

        @dsl.pipeline
        def my_pipeline():
            raise_component()

        task = my_pipeline()
        logged_output = mock_stdout.getvalue()
        # Logs should:
        # - log task failure trace
        # - log pipeline failure
        # - indicate which task the failure came from
        self.assertRegex(
            logged_output,
            r"raise Exception\('Error from raise_component\.'\)",
        )
        self.assertRegex(
            logged_output,
            r"ERROR - Task \x1b\[96m'raise-component'\x1b\[0m finished with status \x1b\[91mFAILURE\x1b\[0m\n",
        )
        self.assertRegex(
            logged_output,
            r"ERROR - Pipeline \x1b\[95m'my-pipeline'\x1b\[0m finished with status \x1b\[91mFAILURE\x1b\[0m\. Inner task failed: \x1b\[96m'raise-component'\x1b\[0m\.\n",
        )
        self.assertEqual(task.outputs, {})

    @mock.patch('sys.stdout', new_callable=stdlib_io.StringIO)
    def test_single_nested_fails_with_raise_on_error_false(self, mock_stdout):
        local.init(local.SubprocessRunner(), raise_on_error=False)

        @dsl.component
        def fail():
            raise Exception('Nested failure!')

        @dsl.pipeline
        def inner_pipeline():
            fail()

        @dsl.pipeline
        def my_pipeline():
            inner_pipeline()

        task = my_pipeline()
        logged_output = mock_stdout.getvalue()
        # Logs should:
        # - log task failure trace
        # - log pipeline failure
        # - indicate which task the failure came from
        self.assertRegex(
            logged_output,
            r"raise Exception\('Nested failure!'\)",
        )
        self.assertRegex(
            logged_output,
            r"ERROR - Task \x1b\[96m'fail'\x1b\[0m finished with status \x1b\[91mFAILURE\x1b\[0m\n",
        )
        self.assertRegex(
            logged_output,
            r"ERROR - Pipeline \x1b\[95m\'my-pipeline\'\x1b\[0m finished with status \x1b\[91mFAILURE\x1b\[0m. Inner task failed: \x1b\[96m\'fail\'\x1b\[0m inside \x1b\[96m\'inner-pipeline\'\x1b\[0m.\n"
        )
        self.assertEqual(task.outputs, {})

    @mock.patch('sys.stdout', new_callable=stdlib_io.StringIO)
    def test_deeply_nested_fails_with_raise_on_error_false(self, mock_stdout):
        local.init(local.SubprocessRunner(), raise_on_error=False)

        @dsl.component
        def fail():
            raise Exception('Nested failure!')

        @dsl.pipeline
        def deep_pipeline():
            fail()

        @dsl.pipeline
        def mid_pipeline():
            deep_pipeline()

        @dsl.pipeline
        def outer_pipeline():
            mid_pipeline()

        task = outer_pipeline()
        logged_output = mock_stdout.getvalue()
        # Logs should:
        # - log task failure trace
        # - log pipeline failure
        # - indicate which task the failure came from
        self.assertRegex(
            logged_output,
            r"raise Exception\('Nested failure!'\)",
        )
        self.assertRegex(
            logged_output,
            r"ERROR - Task \x1b\[96m'fail'\x1b\[0m finished with status \x1b\[91mFAILURE\x1b\[0m\n",
        )
        self.assertRegex(
            logged_output,
            r"ERROR - Pipeline \x1b\[95m'outer-pipeline'\x1b\[0m finished with status \x1b\[91mFAILURE\x1b\[0m\. Inner task failed: \x1b\[96m'fail'\x1b\[0m inside \x1b\[96m'deep-pipeline'\x1b\[0m inside \x1b\[96m'mid-pipeline'\x1b\[0m\.\n",
        )
        self.assertEqual(task.outputs, {})

    def test_fstring_python_component(self):
        local.init(runner=local.SubprocessRunner())

        @dsl.component
        def identity(string: str) -> str:
            return string

        @dsl.pipeline
        def my_pipeline(string: str = 'baz') -> str:
            op1 = identity(string=f'bar-{string}')
            op2 = identity(string=f'foo-{op1.output}')
            return op2.output

        task = my_pipeline()
        self.assertEqual(task.output, 'foo-bar-baz')


class TestFstringContainerComponent(
        testing_utilities.LocalRunnerEnvironmentTestCase):

    @classmethod
    def setUpClass(cls):
        from kfp.local import subprocess_task_handler

        # Temporarily removing these these validation calls is useful hack to
        # test a ContainerComponent outside of a container.
        # We do this here because we only want to test the very specific
        # f-string logic in container components without the presence of
        # Docker in the test environment.
        cls.original_validate_image = subprocess_task_handler.SubprocessTaskHandler.validate_image
        subprocess_task_handler.SubprocessTaskHandler.validate_image = lambda slf, image: None

        cls.original_validate_not_container_component = subprocess_task_handler.SubprocessTaskHandler.validate_not_container_component
        subprocess_task_handler.SubprocessTaskHandler.validate_not_container_component = lambda slf, full_command: None

        cls.original_validate_not_containerized_python_component = subprocess_task_handler.SubprocessTaskHandler.validate_not_containerized_python_component
        subprocess_task_handler.SubprocessTaskHandler.validate_not_containerized_python_component = lambda slf, full_command: None

    @classmethod
    def tearDownClass(cls):
        from kfp.local import subprocess_task_handler

        subprocess_task_handler.SubprocessTaskHandler.validate_image = cls.original_validate_image
        subprocess_task_handler.SubprocessTaskHandler.validate_not_container_component = cls.original_validate_not_container_component
        subprocess_task_handler.SubprocessTaskHandler.validate_not_containerized_python_component = cls.original_validate_not_containerized_python_component

    def test_fstring_container_component(self):
        local.init(runner=local.SubprocessRunner())

        @dsl.container_component
        def identity_container(string: str, outpath: dsl.OutputPath(str)):
            return dsl.ContainerSpec(
                image='alpine',
                command=[
                    'sh',
                    '-c',
                    f"""mkdir -p $(dirname {outpath}) && printf '%s' {string} > {outpath}""",
                ])

        @dsl.pipeline
        def my_pipeline(string: str = 'baz') -> str:
            op1 = identity_container(string=f'bar-{string}')
            op2 = identity_container(string=f'foo-{op1.output}')
            return op2.output

        task = my_pipeline()
        self.assertEqual(task.output, 'foo-bar-baz')


if __name__ == '__main__':
    unittest.main()



# Copyright 2023 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from absl.testing import parameterized
from kfp.compiler import compiler_utils
from kfp.dsl import pipeline_channel


class TestAdditionalInputNameForPipelineChannel(parameterized.TestCase):

    @parameterized.parameters(
        {
            'channel':
                pipeline_channel.create_pipeline_channel(
                    name='output1', task_name='task1', channel_type='String'),
            'expected':
                'pipelinechannel--task1-output1',
        },
        {
            'channel':
                pipeline_channel.create_pipeline_channel(
                    name='output1',
                    task_name='task1',
                    channel_type='system.Artifact@0.0.1',
                ),
            'expected':
                'pipelinechannel--task1-output1',
        },
        {
            'channel':
                pipeline_channel.create_pipeline_channel(
                    name='param1', channel_type='String'),
            'expected':
                'pipelinechannel--param1',
        },
    )
    def test_additional_input_name_for_pipeline_channel(self, channel,
                                                        expected):
        self.assertEqual(
            expected,
            compiler_utils.additional_input_name_for_pipeline_channel(channel))

    @parameterized.parameters(
        {
            'data': [{
                'container_spec': {
                    'image_uri':
                        'gcr.io/ml-pipeline/google-cloud-pipeline-components:2.5.0',
                    'command': ['echo'],
                    'args': ['foo']
                },
                'machine_spec': {
                    'machine_type':
                        pipeline_channel.PipelineParameterChannel(
                            name='Output',
                            channel_type='String',
                            task_name='machine-type'),
                    'accelerator_type':
                        pipeline_channel.PipelineParameterChannel(
                            name='Output',
                            channel_type='String',
                            task_name='accelerator-type'),
                    'accelerator_count':
                        1
                },
                'replica_count': 1
            }],
            'old_value':
                '{{channel:task=machine-type;name=Output;type=String;}}',
            'new_value':
                '{{$.inputs.parameters['
                'pipelinechannel--machine-type-Output'
                ']}}',
            'expected': [{
                'container_spec': {
                    'image_uri':
                        'gcr.io/ml-pipeline/google-cloud-pipeline-components:2.5.0',
                    'command': ['echo'],
                    'args': ['foo']
                },
                'machine_spec': {
                    'machine_type':
                        '{{$.inputs.parameters['
                        'pipelinechannel--machine-type-Output'
                        ']}}',
                    'accelerator_type':
                        pipeline_channel.PipelineParameterChannel(
                            name='Output',
                            channel_type='String',
                            task_name='accelerator-type'),
                    'accelerator_count':
                        1
                },
                'replica_count': 1
            }],
        },
        {
            'data': [{
                'first_name':
                    f'my_first_name: {str(pipeline_channel.PipelineParameterChannel(name="Output", channel_type="String", task_name="first_name"))}',
                'last_name':
                    f'my_last_name: {str(pipeline_channel.PipelineParameterChannel(name="Output", channel_type="String", task_name="last_name"))}',
            }],
            'old_value':
                '{{channel:task=first_name;name=Output;type=String;}}',
            'new_value':
                '{{$.inputs.parameters['
                'pipelinechannel--first_name-Output'
                ']}}',
            'expected': [{
                'first_name':
                    'my_first_name: {{$.inputs.parameters[pipelinechannel--first_name-Output]}}',
                'last_name':
                    f'my_last_name: {str(pipeline_channel.PipelineParameterChannel(name="Output", channel_type="String", task_name="last_name"))}',
            }],
        },
        {
            'data': [{
                'project': 'project',
                'location': 'US',
                'job_configuration_query': {
                    'query': 'SELECT * FROM `project.dataset.input_table`',
                    'destinationTable': {
                        'projectId':
                            'project',
                        'datasetId':
                            'dataset',
                        'tableId':
                            f'output_table_{str(pipeline_channel.PipelineParameterChannel(name="Output", channel_type="String", task_name="table_suffix"))}',
                    },
                    'writeDisposition': 'WRITE_TRUNCATE',
                },
            }],
            'old_value':
                '{{channel:task=table_suffix;name=Output;type=String;}}',
            'new_value':
                '{{$.inputs.parameters['
                'pipelinechannel--table_suffix-Output'
                ']}}',
            'expected': [{
                'project': 'project',
                'location': 'US',
                'job_configuration_query': {
                    'query': 'SELECT * FROM `project.dataset.input_table`',
                    'destinationTable': {
                        'projectId':
                            'project',
                        'datasetId':
                            'dataset',
                        'tableId':
                            'output_table_{{$.inputs.parameters[pipelinechannel--table_suffix-Output]}}',
                    },
                    'writeDisposition': 'WRITE_TRUNCATE',
                },
            }],
        },
    )
    def test_recursive_replace_placeholders(self, data, old_value, new_value,
                                            expected):
        self.assertEqual(
            expected,
            compiler_utils.recursive_replace_placeholders(
                data, old_value, new_value))


if __name__ == '__main__':
    unittest.main()



# Copyright 2023 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for testing local execution."""

import datetime
import functools
import os
import pathlib
import tempfile
from typing import Any, Callable, Dict
import unittest
from unittest import mock

from absl.testing import parameterized
from google.protobuf import json_format
from google.protobuf import message
from kfp import components
from kfp import dsl
from kfp.local import config as local_config
from kfp.local import docker_task_handler

_LOCAL_KFP_PACKAGE_PATH = os.path.join(
    os.path.dirname(__file__),
    os.path.pardir,
    os.path.pardir,
)


def modify_volumes_decorator(
        original_method: Callable[..., Any]) -> Callable[..., Any]:

    def wrapper(self, *args, **kwargs) -> Dict[str, Any]:
        original_volumes = original_method(self, *args, **kwargs)
        LOCAL_KFP_VOLUME = {
            _LOCAL_KFP_PACKAGE_PATH: {
                'bind': _LOCAL_KFP_PACKAGE_PATH,
                'mode': 'rw'
            }
        }
        original_volumes.update(LOCAL_KFP_VOLUME)
        return original_volumes

    return wrapper


class LocalRunnerEnvironmentTestCase(parameterized.TestCase):
    """Test class that uses an isolated filesystem and updates the
    dsl.component decorator to install from the local KFP source, rather than
    the latest release."""

    def setUp(self):
        # ENTER: start each test case without an uninitialized environment
        local_config.LocalExecutionConfig.instance = None

        # ENTER: use tempdir for all tests
        self.working_dir = pathlib.Path.cwd()
        self.temp_dir = tempfile.TemporaryDirectory()
        os.chdir(self.temp_dir.name)

        # ENTER: mount KFP dir to enable install from source for docker runner
        self.original_get_volumes_to_mount = docker_task_handler.DockerTaskHandler.get_volumes_to_mount
        docker_task_handler.DockerTaskHandler.get_volumes_to_mount = modify_volumes_decorator(
            docker_task_handler.DockerTaskHandler.get_volumes_to_mount)

    def tearDown(self):
        # EXIT: use tempdir for all tests
        self.temp_dir.cleanup()
        os.chdir(self.working_dir)

        # EXIT: mount KFP dir to enable install from source for docker runner
        docker_task_handler.DockerTaskHandler.get_volumes_to_mount = self.original_get_volumes_to_mount

    @classmethod
    def setUpClass(cls):
        # ENTER: use local KFP package path for subprocess runner
        cls.original_component, dsl.component = dsl.component, functools.partial(
            dsl.component, kfp_package_path=_LOCAL_KFP_PACKAGE_PATH)

    @classmethod
    def tearDownClass(cls):
        # EXIT: use local KFP package path for subprocess runner
        dsl.component = cls.original_component


class MockedDatetimeTestCase(unittest.TestCase):

    def setUp(self):
        # set up patch, cleanup, and start
        patcher = mock.patch('kfp.local.executor_input_utils.datetime.datetime')
        self.addCleanup(patcher.stop)
        self.mock_datetime = patcher.start()

        # set mock return values
        mock_now = mock.MagicMock(
            wraps=datetime.datetime(2023, 10, 10, 13, 32, 59, 420710))
        self.mock_datetime.now.return_value = mock_now
        mock_now.strftime.return_value = '2023-10-10-13-32-59-420710'


def write_proto_to_json_file(
    proto_message: message.Message,
    file_path: str,
) -> None:
    """Writes proto_message to file_path as JSON."""
    json_string = json_format.MessageToJson(proto_message)

    with open(file_path, 'w') as json_file:
        json_file.write(json_string)


def compile_and_load_component(
    base_component: dsl.base_component.BaseComponent,
) -> dsl.yaml_component.YamlComponent:
    """Compiles a component to PipelineSpec and reloads it as a
    YamlComponent."""
    return components.load_component_from_text(
        json_format.MessageToJson(base_component.pipeline_spec))



# Copyright 2023 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""An abstract base class that specifies the interface of a task handler."""
import abc
from typing import List

from kfp.local import config
from kfp.local import status


class ITaskHandler(abc.ABC):
    """Interface for a TaskHandler."""

    def __init__(
        self,
        image: str,
        full_command: List[str],
        pipeline_root: str,
        runner: config.LocalRunnerType,
    ) -> None:
        pass

    @abc.abstractmethod
    def run(self) -> status.Status:
        """Runs the task and returns the status.

        Returns:
            Status.
        """
        pass
